;*********************************************************************
; FAT CONTROLLER ~ MIDI to CV Converter v1.2                         *
;*********************************************************************
; (c) 2009 SM-Sound / Seb Francis                                    *
;*********************************************************************
;
; Changelog
; ---------
; v1.1  2007
; Added support for note off being sent as MIDI Note On message with velocity 0
; Previously only true MIDI Note Off messages were supported
;
; v1.2  22-10-2009
; Added support for changing the AMOUNT setting of CV assignments,
; via MIDI over the control channel.
;
;
;*********************************************************************
;
; Main code
;
;*********************************************************************
;                                                               
;    Connections:
;
;    RA0   <- Noise
;    RA1   -> MUX0
;    RA2   -> MUX1
;    RA3   -> MUX2
;    RA4   Open drain output!
;    RA5   -> MUX3
;
;    RB0   -> Display RS
;    RB1   -> Display R/!W
;    RB2   -> Display E
;    RB3
;    RB4   <- !PGUP
;    RB5   <- !SHIFT 
;    RB6   <- !INCR
;    RB7   <- !DECR
;
;    RC0   Used by Timer1
;    RC1   Used by Timer1
;    RC2   -> DAC !CS
;    RC3   -> DAC SCLK
;    RC4   -> MUXE
;    RC5   -> DAC DIN
;    RC6
;    RC7   <- MIDI IN
;
;    RD0-7 -> Display D0-7
;
;    RE0   -> !Gate 1
;    RE1   -> !Gate 2
;    RE2   -> !Gate 3
;                                                                  
;                                                                     
;*********************************************************************

; Compiler switches
 	#define	ENABLE_DELAYS

;*********************************************************************


	list      p=16f877	; List directive to define processor
	#include <p16f877.inc>
	#include <const.inc>	; Constants + Memory map
	#include <misc.inc>	; Various macros
	
	
	__CONFIG _CP_OFF & _WDT_OFF & _BODEN_ON & _PWRTE_ON & _HS_OSC & _WRT_ENABLE_OFF & _LVP_OFF & _DEBUG_OFF & _CPD_OFF 

; '__CONFIG' directive is used to embed configuration data within .asm file.
; The lables following the directive are located in the respective .inc file.
; See respective data sheet for additional information on configuration word.




;**********************************************************************


	org     0x0000		; Processor reset vector
	clrf    PCLATH		; Ensure page bits are cleared
	goto    Init		; Go to beginning of program


	org     0x0004		; Interrupt vector location
	
	movwf   W_TEMP		; Save W, STATUS, PCLATH and FSR
	swapf	STATUS,w	; Use swapf, not movf so not to disturb Z flag
	movwf	STATUS_TEMP
	movfw   PCLATH
	movwf	PCLATH_TEMP
	movfw	FSR
	movwf	FSR_TEMP
	
	clrf    PCLATH		; Clear page bits
	

; Call the interupt handlers

	clrf	STATUS		; RAM bank 0, IRP 0
	btfsc	INTCON,T0IF
	call	UpdateCV

	clrf	STATUS		; RAM bank 0, IRP 0
	btfsc	PIR1,RCIF
	call	MIDIByteReceived

	clrf	STATUS		; RAM bank 0, IRP 0
	btfsc	INTCON,RBIF
	call	KeyChange

	clrf	STATUS		; RAM bank 0, IRP 0
	btfsc	PIR1,TMR1IF
	call	KeyTimer

; --
	movfw	FSR_TEMP	; Restore W, STATUS, PCLATH and FSR
	movwf	FSR
	movfw	PCLATH_TEMP
	movwf	PCLATH
	swapf	STATUS_TEMP,w	; Use swapf, not movf so not to disturb Z flag
	movwf	STATUS
	swapf	W_TEMP,f	; Use swapf, not movf so not to disturb Z flag
	swapf	W_TEMP,w	; Swap nibbles again to get them correct
InterruptEnd
	retfie			; Return from interrupt


;**********************************************************************


Init

;
; Setup the I/O and SFR
; 

; Bank 0 Setup
	bcf	STATUS,RP0
	bcf	STATUS,RP1
	
	clrf	PORTA		; Set all outputs to off initially
	clrf	PORTB
	clrf	PORTC
	clrf	PORTD
	movlw	b'00000011'
	movwf	PORTE
	
	clrf	PIR1		; Clear all interrupt flags
	clrf	PIR2
	
	movlw	b'00000000'	; A/D clock FOSC/2, Select RA0, A/D Off
	movwf	ADCON0

	movlw	b'00111000'	; TMR1 pre-scale 1:8, TMR1 osc enabled, Internal clock, TMR1 stopped
	movwf	T1CON
	
	movlw	b'10000000'	; USART: Enabled, 8bit reception, Reception (CREN) disabled, Address detection disabled
	movwf	RCSTA
	
	movlw	b'00000000'	; SPI: Disable serial port, Clock idle low, SPI mode @ 5MHz
	movwf	SSPCON
	
	; Initialise general data memory
	
	clrf	MENU_CONFIG_POS	; Initialise position in config menu
	clrf	MENU_ASSIGN_POS	; Initialise position in assignment menu
	movlw	1		; Select CV1 in assignment menu
	movwf	MENU_CURRENT_CV
	clrf	MENU_FLAGS	; Clear menu flags
	
	clrf	MIDI_FLAGS	; Clear MIDI flags
	clrf	MIDI_MSG1	; Clear MIDI status buffer
	
	movlw	low MIDI_MSG_BUFFER	; Set pointers to MIDI ring buffer
	movwf	MIDI_MSG_PROCESS_PTR
	movwf	MIDI_MSG_INCOMING_PTR
	
	
; Bank 1 Setup
	bsf	STATUS,RP0
	bcf	STATUS,RP1
	
	movlw	b'10001110'	; Right justified, RA0 analog, RA1-7 digital, Internal Vrefs
	movwf	ADCON1
	
	movlw	b'00000001'	; Port A (RA0 input)
	movwf	TRISA
	
	movlw	b'11110000'	; Port B (RB4-7 inputs)
	movwf	TRISB
	
	movlw	b'10000000'	; Port C (RC7 input)
	movwf	TRISC
	
	movlw	b'00000000'	; Port D (all outputs)
	movwf	TRISD
	
	movwf	TRISE		; Port E (all outputs)
	
	movlw	b'00001000'	; PORTB pullups enabled, Internal TMR0 clock, Pre-scaler assigned to WDT
	movwf	OPTION_REG	; TMR0 with no pre-scale gives update rate of 51.2us per channel, 768us overall = 1.302 KHz sample rate
				; ISR with no other code (just updating CV output) takes 23us
				; S&H has 51.2-7.6 = 43.6us to settle (needs ~20us)
	
	movlw	b'01101000'	; Interupts: Global off, Peripheral on, TMR0 on, External off, RB on
	movwf	INTCON
	
	movlw	b'00100001'	; Peripheral interrupts 1: USART on, TMR1 on
	movwf	PIE1
	
	movlw	b'00000000'	; Peripheral interrupts 2: All off
	movwf	PIE2
	
	movlw	b'00000000'	; USART: Transmit disabled, Asynchronous, Low Baud Rate
	movwf	TXSTA
	
	movlw	D'9'		; USART: Baud rate 31,250
	movwf	SPBRG
	
	movlw	b'01000000'	; SPI: Transmit on rising edge
	movwf	SSPSTAT

	
	; Initialise general data memory
	
	clrf	NOTE1_AMT_LL
	clrf	NOTE2_AMT_LL
	clrf	NOTE3_AMT_LL
	movlw	0xFE
	movwf	NOTE1_AMT_L
	movwf	NOTE2_AMT_L
	movwf	NOTE3_AMT_L
	movlw	0x1F
	movwf	NOTE1_AMT_H
	movwf	NOTE2_AMT_H
	movwf	NOTE3_AMT_H
	clrf	BEND1_AMT_L
	clrf	BEND1_AMT_H
	clrf	BEND2_AMT_L
	clrf	BEND2_AMT_H
	clrf	BEND3_AMT_L
	clrf	BEND3_AMT_H
	
	movlw	0x11		; Set pointer to top of note + velocity stack
	movwf	NOTE_STACK_PTR

	movlw	1
	movwf	NEXT_CV		; Update CV1 next
	
	bcf	STATUS,IRP	; Set all CV levels to 0V (0x1FFE)
	movlw	D'15'
	movwf	COUNTER
	movlw	low CV1_VALUE_L
	movwf	FSR
Init_CV_Loop
	movlw	0xFE
	movwf	INDF
	incf	FSR
	movlw	0x1F
	movwf	INDF
	incf	FSR
	decfsz	COUNTER
	goto	Init_CV_Loop


; Bank 2 Setup
	bcf	STATUS,RP0
	bsf	STATUS,RP1
	
	
	; Load settings from EEPROM
	
	bsf	STATUS,RP0		; Bank 3
	bcf	EECON1,EEPGD		; Point to EEPROM data memory
	bcf	STATUS,RP0		; Bank 2
	
	movlw	EEPROM_FLAGS		; First read EEPROM_FLAGS
	movwf	EEADR
	bsf	STATUS,RP0		; Bank 3
	bsf	EECON1,RD		; Start read operation
	bcf	STATUS,RP0		; Bank 2
	movfw	EEDATA
	andlw	0x1
	movwf	TEMP1			; TEMP1 contains EEPROM_VALID in bit0
	
	; If shift key held on power up, override EEPROM_FLAGS and load parameter defaults
	btfss	PORTB,SHIFT
	bsf	TEMP1,0
	
	bsf	STATUS,IRP		; Set IRP for indirect writing to bank 2
	movlw	MENU_CONFIG_LEN
	movwf	COUNTER			; Count MENU_CONFIG_LEN values
	clrf	TEMP2			; TEMP2 contains menu page offset
	
Init_EEPROM_Load_Loop
	GVFT	MaxValues,TEMP2		; Find if value is 16bit
	addlw	0
	snz
	bsf	TEMP1,1			; TEMP1 contains 16bit flag in bit1
	
	GVFT	ValueAddresses,TEMP2	; Get value address
	movwf	FSR			; Point INDF to value
	
	btfsc	TEMP1,EEPROM_VALID	; Need to read from EEPROM, or initialise with default value?
	goto	Init_GetDefaultValue
	
	movwf	EEADR			; Read byte from EEPROM
	bsf	STATUS,RP0		; Bank 3
	bsf	EECON1,RD		; Start read operation
	bcf	STATUS,RP0		; Bank 2
	movfw	EEDATA
Init_GotDefaultValue
	movwf	INDF			; Write value to memory
	
	btfsc	TEMP1,1			; Need to set high byte?
	call	Init_SetHighByte
	
	incf	TEMP2			; Increment page offset
	decfsz	COUNTER			; Decrement and test counter
	goto	Init_EEPROM_Load_Loop
	goto	Init_EEPROM_LoadAssign
	
Init_SetHighByte
	incf	FSR			; Next byte
	movfw	FSR
	
	btfsc	TEMP1,EEPROM_VALID	; Need to read from EEPROM, or initialise with min value?
	goto	Init_GetDefaultValue_H
	
	movwf	EEADR			; Read byte from EEPROM
	bsf	STATUS,RP0		; Bank 3
	bsf	EECON1,RD		; Start read operation
	bcf	STATUS,RP0		; Bank 2
	movfw	EEDATA
Init_GotDefaultValue_H
	movwf	INDF			; Write value to memory
	return


Init_GetDefaultValue
	GVFT	DefaultValues,TEMP2
	goto	Init_GotDefaultValue
	
Init_GetDefaultValue_H
	GVFT	DefaultValues,TEMP2
	xorlw	0xFF			; Test if default = 0xFF
	movlw	0
	jnz	Init_GotDefaultValue_H
	movlw	0xFF			; If so, set MSB to 0xFF
	goto	Init_GotDefaultValue_H


	
Init_EEPROM_LoadAssign
	movlw	low AssignMenuValuesBase ; Start address
	movwf	FSR
	movlw	D'30'			; Count 30
	movwf	COUNTER
Init_EEPROM_Load_Loop1
	
	btfsc	TEMP1,EEPROM_VALID	; Need to read from EEPROM, or initialise with default value?
	goto	Init_Get0
	
	movfw	FSR
	movwf	EEADR			; Read byte from EEPROM
	bsf	STATUS,RP0		; Bank 3
	bsf	EECON1,RD		; Start read operation
	bcf	STATUS,RP0		; Bank 2
	movfw	EEDATA
	
Init_Got0
	movwf	INDF			; Write value to memory
	
	incf	FSR			; Increment address
	decfsz	COUNTER			; Decrement and test counter
	goto	Init_EEPROM_Load_Loop1
	goto	Init_EEPROM_Loaded
	
Init_Get0
	movlw	0
	goto	Init_Got0


Init_EEPROM_Loaded


; Bank 3 Setup
	bsf	STATUS,RP0
	bsf	STATUS,RP1
	
	clrf	PORT1_STEPS_L		; Disable portamento
	clrf	PORT1_STEPS_H
	clrf	PORT2_STEPS_L
	clrf	PORT2_STEPS_H
	clrf	PORT3_STEPS_L
	clrf	PORT3_STEPS_H
	
	clrf	ENV1_STEPS_L		; Disable envelopes
	clrf	ENV1_STEPS_H
	clrf	ENV2_STEPS_L
	clrf	ENV2_STEPS_H
	
;--
	clrf	STATUS		; Bank 0 Select
	
	call	LCD_Init	; Display init
	call	UpdateDisplay	; Write initial page contents

	bsf	PORTC,2		; Set !CS high
	bsf	SSPCON,SSPEN	; SPI: Enable serial port
	bsf	RCSTA,CREN	; Enable MIDI receive
	bcf	INTCON,T0IF	; Reset TMR0 flag
	bsf	INTCON,GIE	; Enable interrupts


;#######################################################################

; RAM Bank 0 !
Main
	btfsc	MENU_FLAGS,SAVE_CONFIG_GO	; Check if config needs saving
	call	SaveConfigGo
	
Main_InnerLoop
	bsf	PCLATH,3	; Select program bank 3
	bsf	PCLATH,4
	btfsc	MENU_FLAGS,ENV1_CALC_PENDING	; Need to make envelope calculations?
	call	CalculateEnv1
	btfsc	MENU_FLAGS,ENV2_CALC_PENDING
	call	CalculateEnv2
	clrf	PCLATH		; Select program bank 0

	movfw	MIDI_MSG_PROCESS_PTR	; Check if MIDI buffer has any messages
	subwf	MIDI_MSG_INCOMING_PTR,w
	jz	Main
	
	movlw	high ProcessMIDI
	movwf	PCLATH			; Select program bank 2, and ready PCLATH for jump table
	call	ProcessMIDI
	clrf	PCLATH
	goto	Main_InnerLoop
	
	
; *******************************************************************
; *******************************************************************
; *******************************************************************
	
	
SaveConfigGo
	bsf	STATUS,RP0	; RAM Bank 3
	bsf	STATUS,RP1
	
	movlw	EEPROM_Top - EEPROM_ConfigMenuValuesBase + 1
	movwf	MTEMP2		; Initialise counter with number of bytes to write
	
	bsf	STATUS,IRP	; Set IRP for indirect reading from bank 2
	movlw	low ConfigMenuValuesBase	; Start address
	movwf	FSR
	
SaveConfigGo_Loop
	
	; Check for MIDI messages in between writing each byte to EEPROM
	bcf	STATUS,RP0	; RAM Bank 0
	bcf	STATUS,RP1
	call	SaveConfigGo_CheckForMIDI
	bsf	STATUS,RP0	; RAM Bank 3
	bsf	STATUS,RP1

	movfw	INDF		; Copy INDF to EEPROM
	movwf	MTEMP1
	call	WriteByteToEEPROM
	
	incf	FSR		; Increment pointer
	decfsz	MTEMP2		; Decrement and test counter
	goto	SaveConfigGo_Loop
	
	; Finally set EEPROM_VALID flag
	clrf	MTEMP1
	clrf	FSR
	call	WriteByteToEEPROM
	
	
	bcf	STATUS,RP0	; RAM Bank 0
	bcf	STATUS,RP1
	
	; All done, so clear 'save config' mode flags
	bcf	MENU_FLAGS,SAVE_CONFIG
	bcf	MENU_FLAGS,SAVE_CONFIG_GO
	return
	
SaveConfigGo_CheckForMIDI
	movfw	MIDI_MSG_PROCESS_PTR	; Check if buffer has any messages
	subwf	MIDI_MSG_INCOMING_PTR,w
	rz

	; Save FSR + MTEMP1 + MTEMP2 before processing MIDI message
	movfw	FSR
	movwf	EEPROM_FSR_SAVE
	movfw	MTEMP1
	movwf	EEPROM_MTEMP1_SAVE
	movfw	MTEMP2
	movwf	EEPROM_MTEMP2_SAVE
	
	movlw	high ProcessMIDI
	movwf	PCLATH		; Select program bank 2, and ready PCLATH for jump table
	call	ProcessMIDI
	clrf	PCLATH
	
	bsf	STATUS,IRP	; Re-set IRP for indirect reading from bank 2
	; Restore FSR + MTEMP1 + MTEMP2
	movfw	EEPROM_FSR_SAVE
	movwf	FSR
	movfw	EEPROM_MTEMP1_SAVE
	movwf	MTEMP1
	movfw	EEPROM_MTEMP2_SAVE
	movwf	MTEMP2
	
	goto	SaveConfigGo_CheckForMIDI
	
	
; Must be call with RAM Bank 3 selected
; Returns with RAM Bank 3 selected
; FSR = Address to write to
; MTEMP1 = Value to write
WriteByteToEEPROM
	; Write byte to EEPROM
	btfsc	EECON1,WR	; Wait for write to finish
	goto	$-1
	
	bcf	STATUS,RP0	; RAM Bank 2
	
	movfw	FSR		; Address to write to
	movwf	EEADR
	movfw	MTEMP1		; Value to write
	movwf	EEDATA
	
	bsf	STATUS,RP0	; RAM Bank 3
	
	bcf	EECON1,EEPGD	; Point to data memory
	bsf	EECON1,WREN	; Enable writes
	bcf	INTCON,GIE	; Disable interrupts
	
	movlw	0x55		; Write 0x55 to EECON2
	movwf	EECON2
	movlw	0xAA		; Write 0xAA to EECON2
	movwf	EECON2
	bsf	EECON1,WR	; Set WR bit (starts write operation)
	
	bsf	INTCON,GIE	; Enable interrupts
	bcf	EECON1,WREN	; Disable writes
	
	return
	

;#######################################################################



; --------------------------------


UpdateCV
	bsf	STATUS,RP0	; Select RAM bank 1
	
	; Jump to:
	; CV1 (0001)  -> 0x0800 (00001 000  0 0000000)
	; CV2 (0010)  -> 0x0880 (00001 000  1 0000000)
	; CV3 (0011)  -> 0x0900 (00001 001  0 0000000)
	; CV4 (0100)  -> 0x0980 (00001 001  1 0000000)
	; CV5 (0101)  -> 0x0A00 (00001 010  0 0000000)
	; CV6 (0110)  -> 0x0A80 (00001 010  1 0000000)
	; CV7 (0111)  -> 0x0B00 (00001 011  0 0000000)
	; CV8 (1000)  -> 0x0B80 (00001 011  1 0000000)
	; CV9 (1001)  -> 0x0C00 (00001 100  0 0000000)
	; CV10 (1010) -> 0x0C80 (00001 100  1 0000000)
	; CV11 (1011) -> 0x0D00 (00001 101  0 0000000)
	; CV12 (1100) -> 0x0D80 (00001 101  1 0000000)
	; CV13 (1101) -> 0x0E00 (00001 110  0 0000000)
	; CV14 (1110) -> 0x0E80 (00001 110  1 0000000)
	; CV15 (1111) -> 0x0F00 (00001 111  0 0000000)
	
	decf	NEXT_CV,w		; Make CV pointer 0 based
	movwf	PCLATH			; Set into PCLATH
	rrf	PCLATH			; Rotate right
	bsf	PCLATH,3		; Set bit 3
	clrf	TEMP1
	btfsc	STATUS,C		; Carry last bit into w
	bsf	TEMP1,7
	
	movfw	TEMP1
	movwf	PCL			; Jump to routine

; --------------------------------


PointToNextIncomingBufferByte	macro
	incf	FSR			; Increment pointer
	btfss	FSR,4			; Check not wrapped
	goto	$+3
	movlw	low MIDI_MSG_BUFFER	; Wrap pointer
	movwf	FSR
	
	movfw	FSR			; Check for ring buffer overrun
	subwf	MIDI_MSG_PROCESS_PTR,w
	jz	MIDIByteReceived_RingBuffOver
endm

MIDIByteReceived_RingBuffOver
	clrf	MIDI_MSG1		; Clear status byte and flags
	clrf	MIDI_FLAGS
	goto	MIDIByteReceived_TooSlowError


; RAM bank 0
MIDIByteReceived

	movfw	RCSTA			; Read and save and error code
	movwf	TEMP2
	
	movfw	RCREG			; Read MIDI byte
	movwf	TEMP1			; Save in TEMP1
	
	btfsc	TEMP2,OERR		; Check for overrun error
	goto	MIDIByteReceived_OverrunError
	btfsc	TEMP2,FERR		; Check for framing error
	goto	MIDIByteReceived_FramingError
	
	
	btfss	TEMP1,7			; Header or data byte?
	goto	MIDIByteReceived_DataByte
	
; header byte
	movlw	b'11110000'		; System message
	andwf	TEMP1,w
	sublw	b'11110000'
	jz	MIDIByteReceived_System
	
	; Voice message
	clrf	MIDI_FLAGS		; Clear flags
	
	movfw	TEMP1			; Store status byte
	movwf	MIDI_MSG1
	
	movlw	b'11100000'		; 2 byte messages (Program Change + Aftertouch)
	andwf	TEMP1,w
	sublw	b'11000000'
	sz				; If not 2 byte message, set flag
	bsf	MIDI_FLAGS,MSG_LEN
	
	return
	
MIDIByteReceived_System
	btfss	TEMP1,3			; System common message?
	goto	MIDIByteReceived_SystemCommon
	
	; System realtime message
	movfw	MIDI_MSG_INCOMING_PTR
	movwf	FSR
	PointToNextIncomingBufferByte	; Point INDF to next byte of ring buffer
	movfw	FSR
	movwf	MIDI_MSG_INCOMING_PTR
	
	movfw	TEMP1			; Write byte
	movwf	INDF
	
	return
	
MIDIByteReceived_SystemCommon
	clrf	MIDI_MSG1		; Clear status buffer
	
	return				; Ignore system common messages
	

MIDIByteReceived_DataByte
	movfw	MIDI_MSG1		; Drop byte if status is 0
	rz
	
	btfsc	MIDI_FLAGS,MSG_LEN	; Test if expecting 3 byte message
	goto	MIDIByteReceived_DataByte_3byte
	
	; 2 byte message
	movfw	MIDI_MSG_INCOMING_PTR
	movwf	FSR
	PointToNextIncomingBufferByte	; Point INDF to next byte of ring buffer
	
	movfw	MIDI_MSG1		; Write status byte
	movwf	INDF
	
	PointToNextIncomingBufferByte	; Point INDF to next byte of ring buffer
	movfw	FSR
	movwf	MIDI_MSG_INCOMING_PTR
	
	movfw	TEMP1			; Write data byte
	movwf	INDF
	
	return
	
MIDIByteReceived_DataByte_3byte
	btfsc	MIDI_FLAGS,WHICH_BYTE	; Test if this is the 2nd data byte
	goto	MIDIByteReceived_DataByte_3byte2
	
	; 3 byte message, 1st data byte
	movfw	TEMP1			; Store data byte
	movwf	MIDI_MSG2
	
	bsf	MIDI_FLAGS,WHICH_BYTE	; Set flag to expect 2nd data byte
	
	return
	
MIDIByteReceived_DataByte_3byte2
	; 3 byte message, 2nd data byte
	movfw	MIDI_MSG_INCOMING_PTR
	movwf	FSR
	PointToNextIncomingBufferByte	; Point INDF to next byte of ring buffer
	
	movfw	MIDI_MSG1		; Write status byte
	movwf	INDF
	
	PointToNextIncomingBufferByte	; Point INDF to next byte of ring buffer
	
	movfw	MIDI_MSG2		; Write 1st data byte
	movwf	INDF
	
	PointToNextIncomingBufferByte	; Point INDF to next byte of ring buffer
	movfw	FSR
	movwf	MIDI_MSG_INCOMING_PTR
	
	movfw	TEMP1			; Write 2nd data byte
	movwf	INDF
	
	bcf	MIDI_FLAGS,WHICH_BYTE	; Set flag to expect 1st data byte
	
	return
	
	
MIDIByteReceived_OverrunError
	bcf	RCSTA,CREN		; Reset receive logic / clear error flag
	bsf	RCSTA,CREN
	
	movlw	sOVERRUN / 8		; Write error message and return
	call	LCD_WriteTopRowString
	movlw	sERROR / 8
	goto	LCD_WriteBottomRowString
	
MIDIByteReceived_FramingError
	bcf	RCSTA,CREN		; Reset receive logic / clear error flag
	bsf	RCSTA,CREN
	
	movlw	sFRAMING / 8		; Write error message and return
	call	LCD_WriteTopRowString
	movlw	sERROR / 8
	goto	LCD_WriteBottomRowString
	
MIDIByteReceived_TooSlowError
	movlw	sTOO_SLOW / 8		; Write error message and return
	call	LCD_WriteTopRowString
	movlw	sERROR / 8
	goto	LCD_WriteBottomRowString
	

; --------------------------------
	
	
KeyChange
	bcf	T1CON,TMR1ON		; Stop TMR1
	bcf	MENU_FLAGS,KEY_HELD	; Clear key held flag
	bcf	MENU_FLAGS,BIG_STEP	; Clear big step flag
	
IFDEF	ENABLE_DELAYS
	movlw	0xE7		; Set TMR1 to 10ms = 0x10000 - 10e-3/(8 * 200e-9) = 0xE796
	movwf	TMR1H		; This gives time for switch to stop "bouncing" (max bounce time is specified as 5ms)
	movlw	0xB5
	movwf	TMR1L
ELSE
	movlw	0xFF
	movwf	TMR1H
	movlw	0xFC
	movwf	TMR1L
ENDIF
	movfw	PORTB		; Read from PORTB
	bcf	INTCON,RBIF	; Clear interrupt bit
	bsf	T1CON,TMR1ON	; Start TMR1
	return

	
KeyTimer
	bcf	T1CON,TMR1ON	; Stop TMR1
	
	btfsc	MENU_FLAGS,KEY_HELD	; Check if key held
	goto	KeyTimer_Held
	
	bsf	MENU_FLAGS,KEY_HELD	; Set key held flag
	
IFDEF	ENABLE_DELAYS
	movlw	0x67		; Set TMR1 .. 1/16 second  = 65536 - 1/(16 * 8 * 200e-9)  = 0x6769 ~= 0x6700  (6)
	movwf	TMR1H		;             1/32 second  = 65536 - 1/(32 * 8 * 200e-9)  = 0xB3B4 ~= 0xB400  (5)
	movlw	0x69		;             1/64 second  = 65536 - 1/(64 * 8 * 200e-9)  = 0xD9DA ~= 0xDA00  (4)
	movwf	TMR1L		;             1/128 second = 65536 - 1/(128 * 8 * 200e-9) = 0xECED ~= 0xED00  (3)
				;             1/256 second = 65536 - 1/(256 * 8 * 200e-9) = 0xF676 ~= 0xF600  (2)
				;             1/512 second = 65536 - 1/(512 * 8 * 200e-9) = 0xFB3B ~= 0xFB00  (1)
ELSE
	movlw	0xFF
	movwf	TMR1H
	movlw	0xFC
	movwf	TMR1L
ENDIF
	movlw	0x06		; Initially set TMR1 length to 1/16 second
	movwf	TMR1_LENGTH
	movlw	0x04		; Initially set post-scale count to 4 (halves each step)
	movwf	TMR1_COUNT
	movwf	TMR1_COUNT_SAVE
	movlw	0x02		; Initially 2 iterations at current count rate (doubles each step)
	movwf	TMR1_NUM_COUNTS
	movwf	TMR1_NUM_COUNTS_SAVE
	
	
	btfsc	MENU_FLAGS,SAVE_CONFIG	; 'save config' mode?
	goto	KeyTimer_SaveConfig

	
	btfss	PORTB,SHIFT	; Shift-key down?
	goto	KeyTimer_Shift
	
	
	btfss	PORTB,PGUP	; Page up
	call	PageUp
	
	btfss	PORTB,INCR	; Increment value
	goto	KeyTimer_IV
	
	btfss	PORTB,DECR	; Decrement value
	call	DecrementValue
	
KeyTimer_IV_done
	bcf	PIR1,TMR1IF	; Reset TMR1 flag
	
	comf	PORTB,w		; If incr or decr key is down, start TMR1
	andlw	b'11000000'
	sz
	bsf	T1CON,TMR1ON
	return
	
KeyTimer_IV
	btfss	PORTB,DECR	; Incr + Decr pressed?
	goto	KeyTimer_Both
	
	call	IncrementValue
	goto	KeyTimer_IV_done
	
KeyTimer_Both
	call	ResetValue
	goto	KeyTimer_IV_done
	
			
KeyTimer_Shift
	btfss	PORTB,PGUP	; Page down
	call	PageDown
	
	btfss	PORTB,INCR	; Switch between config and assigment menus
	call	SwitchMenu
	
	btfss	PORTB,DECR	; Switch to 'save config' mode - suspends all other key operations (INCR to confirm, DECR to cancel)
	call	SaveConfigModeOn

	bcf	PIR1,TMR1IF	; Reset TMR1 flag
	
	comf	PORTB,w		; If incr or decr key is down, start TMR1
	andlw	b'11000000'
	sz
	bsf	T1CON,TMR1ON
	return
	
	
KeyTimer_SaveConfig
	btfsc	MENU_FLAGS,SAVE_CONFIG_GO	; Return if save already active
	return

	btfss	PORTB,SHIFT	; Return if shift-key still down
	return
	
	btfss	PORTB,DECR	; Save cancelled
	call	SaveConfigModeOff
	
	btfss	PORTB,INCR	; Save confirmed
	call	SaveConfigModeGo
	
	return
	
	
; --------------------------------

; Table of TMR1 length settings

TMR1Settings
	retlw	0x00
	retlw	0xFB	; 1/512s
	retlw	0xF6	; 1/256s
	retlw	0xED	; 1/128s
	retlw	0xDA	; 1/64s
	retlw	0xB4	; 1/32s
	retlw	0x67	; 1/16s
	
	
KeyTimer_Held
	GVFT	TMR1Settings,TMR1_LENGTH	; Get TMR1 length

IFDEF	ENABLE_DELAYS
	movwf	TMR1H		; Set TMR1
	clrf	TMR1L
ELSE
	movlw	0xFF
	movwf	TMR1H
	movlw	0xFC
	movwf	TMR1L
ENDIF
	bcf	PIR1,TMR1IF	; Reset TMR1 flag
	bsf	T1CON,TMR1ON	; Start TMR1
	
	decfsz	TMR1_COUNT	; Wait for count TMR1 interrupts
	return
	
	
	btfss	PORTB,INCR	; Increment value
	call	IncrementValue
	
	btfss	PORTB,DECR	; Decrement value
	call	DecrementValue
	
	
	movfw	TMR1_COUNT_SAVE	; Reset counter
	movwf	TMR1_COUNT
	
	decfsz	TMR1_NUM_COUNTS	; Iterate at this count rate n times
	return
	
	addlw	0			; Clear carry flag
	rrf	TMR1_COUNT_SAVE		; Half count rate
	movfw	TMR1_COUNT_SAVE
	movwf	TMR1_COUNT
	
	sz			; If count rate gets to 0, set back to 1, decrement TMR1 length pointer
	goto	KeyTimer_Held_1
	
	incf	TMR1_COUNT_SAVE
	incf	TMR1_COUNT
	decf	TMR1_LENGTH
	jz	BigStepMode	; If TMR1 length pointer gets to 0, switch to big step mode
	
KeyTimer_Held_1
	addlw	0			; Clear carry flag
	rlf	TMR1_NUM_COUNTS_SAVE	; Double number of iterations at count rate
	movfw	TMR1_NUM_COUNTS_SAVE
	movwf	TMR1_NUM_COUNTS
	
	return
	
	
BigStepMode
	btfss	MENU_FLAGS,BIG_STEP	; If not already in big step mode, reset key repeat acceleration
	goto	BigStepMode_Switch
	incf	TMR1_LENGTH		; Otherwise just set TMR1_LENGTH back to 1 and return
	return


	; Reset key repeat acceleration to initial state and increase incr/decr step to 100
BigStepMode_Switch
	
	bcf	T1CON,TMR1ON	; Stop TMR1
	
IFDEF	ENABLE_DELAYS
	movlw	0x67		; Set TMR1 .. 1/16 second  = 65536 - 1/(16 * 8 * 200e-9)  = 0x6769 ~= 0x6700  (6)
	movwf	TMR1H		;             1/32 second  = 65536 - 1/(32 * 8 * 200e-9)  = 0xB3B4 ~= 0xB400  (5)
	movlw	0x69		;             1/64 second  = 65536 - 1/(64 * 8 * 200e-9)  = 0xD9DA ~= 0xDA00  (4)
	movwf	TMR1L		;             1/128 second = 65536 - 1/(128 * 8 * 200e-9) = 0xECED ~= 0xED00  (3)
				;             1/256 second = 65536 - 1/(256 * 8 * 200e-9) = 0xF676 ~= 0xF600  (2)
				;             1/512 second = 65536 - 1/(512 * 8 * 200e-9) = 0xFB3B ~= 0xFB00  (1)
ELSE
	movlw	0xFF
	movwf	TMR1H
	movlw	0xFC
	movwf	TMR1L
ENDIF
	movlw	0x06		; Initially set TMR1 length to 1/16 second
	movwf	TMR1_LENGTH
	movlw	0x04		; Initially set post-scale count to 4 (halves each step)
	movwf	TMR1_COUNT
	movwf	TMR1_COUNT_SAVE
	movlw	0x02		; Initially 2 iterations at current count rate (doubles each step)
	movwf	TMR1_NUM_COUNTS
	movwf	TMR1_NUM_COUNTS_SAVE
	
	bsf	MENU_FLAGS,BIG_STEP	; Set big step flag
	
	bcf	PIR1,TMR1IF	; Reset TMR1 flag
	bsf	T1CON,TMR1ON	; Start TMR1
	return
	
	
;#######################################################################

; LCD Functions
#include <lcd.inc>

; Menu Functions
#include <menu.inc>

	
;#######################################################################


	; Program bank 1

	org	0x0800
; RAM bank 1
UpdateCV1
	call	UpdateCVFromRAM
	
	; Calc LFO1
	
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,T0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return
	
; 	bsf	PCLATH,3
; 	call	GenerateRandom_byte
; GotRandom
; 	goto $-1
	
	
GenerateRandom_nybble	macro  buffer, b0, b1, b2, b3
	clrw

	addlw	1		; RND_SHIFT2,4 -> carry flag
	btfsc	RND_SHIFT2,4
	addlw	0xFF
	
	rlf	RND_SHIFT1	; Rotate RND_SHIFT1 left
	
	bcf	buffer,b0	; Carry -> buffer,b0
	btfsc	STATUS,C
	bsf	buffer,b0
	
	addlw	1		; RND_SHIFT3,3 -> carry flag
	btfsc	RND_SHIFT3,3
	addlw	0xFF
	
	rlf	RND_SHIFT2	; Rotate RND_SHIFT2 left
	
	bcf	buffer,b1	; Carry -> buffer,b1
	btfsc	STATUS,C
	bsf	buffer,b1
	
	addlw	1		; RND_SHIFT4,4 -> carry flag
	btfsc	RND_SHIFT4,4
	addlw	0xFF
	
	rlf	RND_SHIFT3	; Rotate RND_SHIFT3 left
	
	bcf	buffer,b2	; Carry -> buffer,b2
	btfsc	STATUS,C
	bsf	buffer,b2
	
	
	; !(RND_SHIFT4,4 xor RND_SHIFT1,4 xor RND_SHIFT3,4) -> carry flag
	
	movf	RND_SHIFT4,w
	xorwf	RND_SHIFT1,w
	xorwf	RND_SHIFT3,w
	movwf	RND_TEMP
	comf	RND_TEMP
	
	clrw
	addlw	1
	btfsc	RND_TEMP,4
	addlw	0xFF
	
	
	rlf	RND_SHIFT4	; Rotate RND_SHIFT4 left
	
	bcf	buffer,b3	; Carry -> buffer,b3
	btfsc	STATUS,C
	bsf	buffer,b3
endm
	
GenerateRandom_byte

	GenerateRandom_nybble RND_BUFFER,0,1,2,3
	GenerateRandom_nybble RND_BUFFER,4,5,6,7
	
	return
	

;-----------------------------------
	org	0x0880
; RAM bank 1
UpdateCV2
	call	UpdateCVFromRAM

	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,T0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return

;-----------------------------------
	org	0x0900
; RAM bank 1
UpdateCV3
	call	UpdateCVFromRAM
	
	; Calc LFO2
	
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,T0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return

;-----------------------------------
	org	0x0980
; RAM bank 1
UpdateCV4
	call	UpdateCVFromRAM
	
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,T0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return
	
;-----------------------------------
	org	0x0A00
; RAM bank 1
UpdateCV5
	call	UpdateCVFromRAM
	
	; Calc ENV1
	
	bcf	STATUS,RP0	; Bank 0 Select
	btfsc	MENU_FLAGS,ENV1_CALC_PENDING	; If Env1 calculation pending, do nothing here
	goto	UpdateCV5_End
	
	bsf	STATUS,RP0	; Bank 3 Select
	bsf	STATUS,RP1
	
	movfw	ENV1_STEPS_L		; Check if any steps left to do
	iorwf	ENV1_STEPS_H,w
	jz	UpdateCV5_End
	
	btfsc	ENV_FLAGS,ENV1_NEG	; Test if step is negative
	goto	UpdateCV5_NegStep
	
	; Positive step
	ADD2424	ENV1_AMT_LL,ENV1_STEP_AMT_LL
	movfw	ENV1_REM_AMT_LL		; Any remainder to add?
	iorwf	ENV1_REM_AMT_L,w
	jz	UpdateCV5_DecSteps
	incf24	ENV1_AMT_LL
	decf16	ENV1_REM_AMT_LL
	goto	UpdateCV5_DecSteps
	
UpdateCV5_NegStep
	; Negative step
	SUB2424	ENV1_AMT_LL,ENV1_STEP_AMT_LL
	movfw	ENV1_REM_AMT_LL		; Any remainder to subtract?
	iorwf	ENV1_REM_AMT_L,w
	jz	UpdateCV5_DecSteps
	decf24	ENV1_AMT_LL
	decf16	ENV1_REM_AMT_LL
	
UpdateCV5_DecSteps
	decf16	ENV1_STEPS_L
	
	; ENV1_AMT_H,ENV1_AMT_L => CV value
	bcf	STATUS,IRP	; Indirectly address RAM banks 0-1
	movlw	low (CV1_VALUE_L - 2)
	movwf	FSR
	bcf	STATUS,RP0	; Bank 2 Select
	movfw	vENV1_DST
	bsf	STATUS,RP0	; Bank 3 Select
	addwf	FSR,f
	addwf	FSR,f
	movfw	ENV1_AMT_L
	movwf	INDF
	incf	FSR
	movfw	ENV1_AMT_H
	movwf	INDF
	
	; If end of Attack phase, switch to Decay/Sustain phase
	movfw	ENV1_STEPS_L	; Check if steps hasn't reached 0 yet
	iorwf	ENV1_STEPS_H,w
	jnz	UpdateCV5_End
	
	btfsc	ENV_FLAGS,ENV1_PHASEA	; Check if Attack phase
	goto	UpdateCV5_End
	
	bsf	ENV_FLAGS,ENV1_PHASEA	; Switch to Decay/Sustain phase
	bcf	ENV_FLAGS,ENV1_PHASEB
	
	bcf	STATUS,RP0	; Bank 0 Select
	bcf	STATUS,RP1
	bsf	MENU_FLAGS,ENV1_CALC_PENDING	; Trigger ENV1 calculation
	

UpdateCV5_End
	bsf	STATUS,RP0	; Bank 1 Select
	bcf	STATUS,RP1
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,T0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return

;-----------------------------------
	org	0x0A80
; RAM bank 1
UpdateCV6
	call	UpdateCVFromRAM

	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,T0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return

;-----------------------------------
	org	0x0B00
; RAM bank 1
UpdateCV7
	call	UpdateCVFromRAM
	
	; Calc ENV2
	
	bcf	STATUS,RP0	; Bank 0 Select
	btfsc	MENU_FLAGS,ENV2_CALC_PENDING	; If Env2 calculation pending, do nothing here
	goto	UpdateCV7_End
	
	bsf	STATUS,RP0	; Bank 3 Select
	bsf	STATUS,RP1
	
	movfw	ENV2_STEPS_L		; Check if any steps left to do
	iorwf	ENV2_STEPS_H,w
	jz	UpdateCV7_End
	
	btfsc	ENV_FLAGS,ENV2_NEG	; Test if step is negative
	goto	UpdateCV7_NegStep
	
	; Positive step
	ADD2424	ENV2_AMT_LL,ENV2_STEP_AMT_LL
	movfw	ENV2_REM_AMT_LL		; Any remainder to add?
	iorwf	ENV2_REM_AMT_L,w
	jz	UpdateCV7_DecSteps
	incf24	ENV2_AMT_LL
	decf16	ENV2_REM_AMT_LL
	goto	UpdateCV7_DecSteps
	
UpdateCV7_NegStep
	; Negative step
	SUB2424	ENV2_AMT_LL,ENV2_STEP_AMT_LL
	movfw	ENV2_REM_AMT_LL		; Any remainder to subtract?
	iorwf	ENV2_REM_AMT_L,w
	jz	UpdateCV7_DecSteps
	decf24	ENV2_AMT_LL
	decf16	ENV2_REM_AMT_LL
	
UpdateCV7_DecSteps
	decf16	ENV2_STEPS_L
	
	; ENV2_AMT_H,ENV2_AMT_L => CV value
	bcf	STATUS,IRP	; Indirectly address RAM banks 0-1
	movlw	low (CV1_VALUE_L - 2)
	movwf	FSR
	bcf	STATUS,RP0	; Bank 2 Select
	movfw	vENV2_DST
	bsf	STATUS,RP0	; Bank 3 Select
	addwf	FSR,f
	addwf	FSR,f
	movfw	ENV2_AMT_L
	movwf	INDF
	incf	FSR
	movfw	ENV2_AMT_H
	movwf	INDF
	
	; If end of Attack phase, switch to Decay/Sustain phase
	movfw	ENV2_STEPS_L	; Check if steps hasn't reached 0 yet
	iorwf	ENV2_STEPS_H,w
	jnz	UpdateCV7_End
	
	btfsc	ENV_FLAGS,ENV2_PHASEA	; Check if Attack phase
	goto	UpdateCV7_End
	
	bsf	ENV_FLAGS,ENV2_PHASEA	; Switch to Decay/Sustain phase
	bcf	ENV_FLAGS,ENV2_PHASEB
	
	bcf	STATUS,RP0	; Bank 0 Select
	bcf	STATUS,RP1
	bsf	MENU_FLAGS,ENV2_CALC_PENDING	; Trigger ENV2 calculation
	

UpdateCV7_End
	bsf	STATUS,RP0	; Bank 1 Select
	bcf	STATUS,RP1
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,T0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return

;-----------------------------------
	org	0x0B80
; RAM bank 1
UpdateCV8
	call	UpdateCVFromRAM

	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,T0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return
	
;-----------------------------------
	org	0x0C00
; RAM bank 1
UpdateCV9
	call	UpdateCVFromRAM

	; Calc PORT1
	bsf	STATUS,RP1	; Bank 3 Select
	
	movfw	PORT1_STEPS_L	; Nothing to do if steps=0
	iorwf	PORT1_STEPS_H,w
	jz	UpdateCV9_End
	
	decf16	PORT1_STEPS_L	; Decrement steps
	
	btfsc	PORT_FLAGS,P1_NEG	; Test if step is negative
	goto	UpdateCV9_NegativeStep
	
	ADD2416Uxb13  NOTE1_AMT_LL, PORT1_STEP_AMT_LL
	
	movfw	PORT1_REM_AMT_LL	; If remainder=0 skip next bit
	iorwf	PORT1_REM_AMT_L,w
	jz	UpdateCV9_AddBend
	
	bcf	STATUS,RP1	; Bank 1 Select
	incf24	NOTE1_AMT_LL
	bsf	STATUS,RP1	; Bank 3 Select
	decf16	PORT1_REM_AMT_LL
	goto	UpdateCV9_AddBend
	
UpdateCV9_NegativeStep
	
	SUB2416Uxb13  NOTE1_AMT_LL, PORT1_STEP_AMT_LL
	
	movfw	PORT1_REM_AMT_LL	; If remainder=0 skip next bit
	iorwf	PORT1_REM_AMT_L,w
	jz	UpdateCV9_AddBend
	
	bcf	STATUS,RP1	; Bank 1 Select
	decf24	NOTE1_AMT_LL
	bsf	STATUS,RP1	; Bank 3 Select
	decf16	PORT1_REM_AMT_LL
	
UpdateCV9_AddBend
	bcf	STATUS,RP1	; Bank 1 Select
	
	; Add in bend amount and write CV value
	AddAndCheck14bit  NOTE1_AMT_L, BEND1_AMT_L, CV1_VALUE_L

UpdateCV9_End
	bcf	STATUS,RP1	; Bank 1 Select
	
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,T0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return
	
;-----------------------------------
	org	0x0C80
; RAM bank 1
UpdateCV10
	call	UpdateCVFromRAM

	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,T0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return
	
;-----------------------------------
	org	0x0D00
; RAM bank 1
UpdateCV11
	call	UpdateCVFromRAM
	
	; Calc PORT2
	bsf	STATUS,RP1	; Bank 3 Select
	
	movfw	PORT2_STEPS_L	; Nothing to do if steps=0
	iorwf	PORT2_STEPS_H,w
	jz	UpdateCV11_End
	
	decf16	PORT2_STEPS_L	; Decrement steps
	
	btfsc	PORT_FLAGS,P2_NEG	; Test if step is negative
	goto	UpdateCV11_NegativeStep
	
	ADD2416Uxb13  NOTE2_AMT_LL, PORT2_STEP_AMT_LL
	
	movfw	PORT2_REM_AMT_LL	; If remainder=0 skip next bit
	iorwf	PORT2_REM_AMT_L,w
	jz	UpdateCV11_AddBend
	
	bcf	STATUS,RP1	; Bank 1 Select
	incf24	NOTE2_AMT_LL
	bsf	STATUS,RP1	; Bank 3 Select
	decf16	PORT2_REM_AMT_LL
	goto	UpdateCV11_AddBend
	
UpdateCV11_NegativeStep
	
	SUB2416Uxb13  NOTE2_AMT_LL, PORT2_STEP_AMT_LL
	
	movfw	PORT2_REM_AMT_LL	; If remainder=0 skip next bit
	iorwf	PORT2_REM_AMT_L,w
	jz	UpdateCV11_AddBend
	
	bcf	STATUS,RP1	; Bank 1 Select
	decf24	NOTE2_AMT_LL
	bsf	STATUS,RP1	; Bank 3 Select
	decf16	PORT2_REM_AMT_LL
	
UpdateCV11_AddBend
	bcf	STATUS,RP1	; Bank 1 Select
	
	; Add in bend amount and write CV value
	AddAndCheck14bit  NOTE2_AMT_L, BEND2_AMT_L, CV2_VALUE_L

UpdateCV11_End
	bcf	STATUS,RP1	; Bank 1 Select
	
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,T0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return
	
;-----------------------------------
	org	0x0D80
; RAM bank 1
UpdateCV12
	call	UpdateCVFromRAM

	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,T0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return

;-----------------------------------
	org	0x0E00
; RAM bank 1
UpdateCV13
	call	UpdateCVFromRAM
	
	; Calc PORT3
	bsf	STATUS,RP1	; Bank 3 Select
	
	movfw	PORT3_STEPS_L	; Nothing to do if steps=0
	iorwf	PORT3_STEPS_H,w
	jz	UpdateCV13_End
	
	decf16	PORT3_STEPS_L	; Decrement steps
	
	btfsc	PORT_FLAGS,P3_NEG	; Test if step is negative
	goto	UpdateCV13_NegativeStep
	
	ADD2416Uxb13  NOTE3_AMT_LL, PORT3_STEP_AMT_LL
	
	movfw	PORT3_REM_AMT_LL	; If remainder=0 skip next bit
	iorwf	PORT3_REM_AMT_L,w
	jz	UpdateCV13_AddBend
	
	bcf	STATUS,RP1	; Bank 1 Select
	incf24	NOTE3_AMT_LL
	bsf	STATUS,RP1	; Bank 3 Select
	decf16	PORT3_REM_AMT_LL
	goto	UpdateCV13_AddBend
	
UpdateCV13_NegativeStep
	
	SUB2416Uxb13  NOTE3_AMT_LL, PORT3_STEP_AMT_LL
	
	movfw	PORT3_REM_AMT_LL	; If remainder=0 skip next bit
	iorwf	PORT3_REM_AMT_L,w
	jz	UpdateCV13_AddBend
	
	bcf	STATUS,RP1	; Bank 1 Select
	decf24	NOTE3_AMT_LL
	bsf	STATUS,RP1	; Bank 3 Select
	decf16	PORT3_REM_AMT_LL
	
UpdateCV13_AddBend
	bcf	STATUS,RP1	; Bank 1 Select
	
	; Add in bend amount and write CV value
	AddAndCheck14bit  NOTE3_AMT_L, BEND3_AMT_L, CV3_VALUE_L

UpdateCV13_End
	bcf	STATUS,RP1	; Bank 1 Select
	
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,T0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return
	
;-----------------------------------
	org	0x0E80
; RAM bank 1
UpdateCV14
	call	UpdateCVFromRAM
	
	; Count note-off gate
	movfw	NOTE_OFF_GATE_STEPS_L
	iorwf	NOTE_OFF_GATE_STEPS_H,w
	jz	UpdateCV14_End	; If STEPS=0 do nothing
	
	; Decrement step counter
	decf16	NOTE_OFF_GATE_STEPS_L
	
	movfw	NOTE_OFF_GATE_STEPS_L
	iorwf	NOTE_OFF_GATE_STEPS_H,w
	jnz	UpdateCV14_End	; If STEPS!=0 do nothing more
	
	; Set note off gate low
	bcf	STATUS,RP0	; Bank 0 Select
	bsf	PORTE,GATE2	; Note off gate low
	
	bsf	PCLATH,4	; Select program bank 3
	movlw	2
	call	CheckAndTriggerEnvReleases
	bcf	PCLATH,4	; Select program bank 1
	
	bsf	STATUS,RP0	; Bank 1 Select

UpdateCV14_End
	incf	NEXT_CV		; Point to next CV
	bcf	INTCON,T0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return
	
;-----------------------------------
	org	0x0F00
; RAM bank 1
UpdateCV15
	call UpdateCVFromRAM

	clrf	NEXT_CV		; Point to CV1
	incf	NEXT_CV
	bcf	INTCON,T0IF	; Reset TMR0 flag
	clrf    PCLATH		; Clear page bits and return
	return
	
;-----------------------------------

; RAM bank 1
UpdateCVFromRAM
	bcf	STATUS,IRP	; Indirectly address RAM bank 0-1
	
	movlw	low (CV1_VALUE_L - 2)	; Base address
	movwf	FSR
	movfw	NEXT_CV		; Add 2xoffset to base address
	addwf	FSR,f
	addwf	FSR,f
	
	movwf	TEMP3		; Save offset in bank 0 accessible RAM
	
	bcf	STATUS,RP0	; Select RAM bank 0
	
	movfw	INDF		; Get LSB
	movwf	TEMP1
	incf	FSR		; Get MSB
	movfw	INDF
	movwf	TEMP2
	
	addlw	0		; Clear carry flag
	rlf	TEMP1		; Shift bits 2 places to left
	rlf	TEMP2
	rlf	TEMP1
	rlf	TEMP2,w
	
	; Set DAC to voltage
	bcf	PORTC,DACCS	; Set !CS low
	movwf	SSPBUF		; Transmit MSB
	nop2			; Wait 9 cycles before sending next byte
	nop2
	nop2
	nop
	bcf	MUXE		; While waiting: Turn off MUX (takes max 300ns)  ; Maybe move this to earlier to eliminate digital feedthrough from the SPI signal???
	movfw	TEMP1		; While waiting: Load LSB into w
	movwf	SSPBUF		; Transmit LSB and sub-bits
	nop2			; Wait 8 cycles
	rlf	TEMP3		; While waiting: Shift MUX bits to left
	btfsc	TEMP3,4		; While waiting: Copy bit 4 to left
	bsf	TEMP3,5
	movfw	TEMP3
	movwf	PORTA		; While waiting: Select MUX channel
	nop
	bsf	PORTC,DACCS	; Set !CS high (loads DAC)
	
	wait_us 5		; Wait 5us for DAC and opamps to settle
				; Maybe less wait with faster opamp?
	
	bsf	MUXE		; Turn on MUX (takes typ 105ns, max 600ns)
	
	; S&H now has until next call of UpdateCVFromRAM to settle

	bsf	STATUS,RP0	; Select RAM bank 1
	
	return


;#######################################################################


	; Program bank 2
	
	org	0x1000
	
	; Implemented messages:
	;
	; 1000cccc nnnnnnnn vvvvvvvv  Note Off
	; 1001cccc nnnnnnnn vvvvvvvv  Note On
	; 1011cccc cccccccc vvvvvvvv  Control Change
	; 1101cccc pppppppp           Aftertouch
	; 1110cccc llllllll hhhhhhhh  Pitch Bend
	;
	; 11111000                    Clock Pulse
	; 11111010                    Start
	; 11111011                    Continue
	; 11111100                    Stop

; RAM Bank 0
PointToNextProcessBufferByte
	incf	MIDI_MSG_PROCESS_PTR	; Increment pointer
	btfss	MIDI_MSG_PROCESS_PTR,4	; Check not wrapped
	goto	$+3
	movlw	low MIDI_MSG_BUFFER	; Wrap pointer
	movwf	MIDI_MSG_PROCESS_PTR
	return
	

; PBA in w, PB in LSB:MTEMP1 MSB:INDF
; Bend amount returned in TMP_NOTE_VALUE_H,TMP_NOTE_VALUE
; RAM Bank 1
PitchBendToDacOffset
	movwf	MTEMP2		; PBA->MTEMP2
	movwf	BARGB0		; PBA->BARGB0

	addlw	0		; Check if PBA=0, if so calc=0
	snz
	goto	$+3
	
	btfss	MTEMP2,7	; Check if PBA>127 (i.e.=0xFF), if so calc=0
	goto	$+4
	
	clrf	TMP_NOTE_VALUE
	clrf	TMP_NOTE_VALUE_H
	return			; Return


	; Pitch bend -> DAC offset
	; (0x5B * PBA) - (PB * PBA * 0x5B >> 13)
	
	; 0x5B * PBA -> TMP_NOTE_VALUE_H, TMP_NOTE_VALUE
	movlw	0x5B
	movwf	AARGB0
	call	FXM0707U
	movfw	AARGB0
	movwf	TMP_NOTE_VALUE_H
	movfw	AARGB1
	movwf	TMP_NOTE_VALUE
	
	; PB * PBA (max value = 0x3FFF * 0x7F = 0x1FBF81 = 21bits)
	movfw	MTEMP2
	movwf	BARGB0
	movfw	INDF
	movwf	AARGB0
	movfw	MTEMP1
	movwf	AARGB1
	call	FXM1507U
	
	; >> 5
	rrf	AARGB0
	rrf	AARGB1
	rrf	AARGB2
	bcf	STATUS,C
	rrf	AARGB0
	rrf	AARGB1
	rrf	AARGB2
	bcf	STATUS,C
	rrf	AARGB0
	rrf	AARGB1
	rrf	AARGB2
	bcf	STATUS,C
	rrf	AARGB0
	rrf	AARGB1
	rrf	AARGB2
	bcf	STATUS,C
	rrf	AARGB0
	rrf	AARGB1,w
	rrf	AARGB2
	movwf	AARGB0
	movfw	AARGB2
	movwf	AARGB1

	; * 0x5B
	movlw	0x5B
	movwf	BARGB0
	call	FXM1507U
	
	; >> 8 = use AARGB0, AARGB1
	
	; (0x5B * PBA) - (PB * PBA * 0x5B >> 13)
	SUB1616	TMP_NOTE_VALUE, AARGB1
	
	return

;***********************************************************************
;***********************************************************************

; Parse MIDI message here
; RAM Bank 0
ProcessMIDI
	call	PointToNextProcessBufferByte
	
	bcf	STATUS,IRP	; Indirectly address banks 0-1
	movfw	MIDI_MSG_PROCESS_PTR
	movwf	FSR
	
	movfw	INDF		; Get and save MIDI channel
	andlw	b'00001111'
	movwf	MTEMP1
	
	rrf	INDF,w		; Get message
	movwf	MTEMP2
	rrf	MTEMP2
	rrf	MTEMP2
	rrf	MTEMP2,w
	andlw	b'00000111'
	
	addwf	PCL		; Jump to handler (PCLATH already set before call to ProcessMIDI)
	goto	ProcessMIDI_NoteOff
	goto	ProcessMIDI_NoteOn
	goto	ProcessMIDI_PolyAftertouch
	goto	ProcessMIDI_ControlChange
	goto	ProcessMIDI_ProgramChange
	goto	ProcessMIDI_Aftertouch
	goto	ProcessMIDI_PitchBend
	goto	ProcessMIDI_System


;***********************************************************************
;***********************************************************************
	
; RAM Bank 0
ProcessMIDI_NoteOff
	call	PointToNextProcessBufferByte	; Point FSR to note value
	movfw	MIDI_MSG_PROCESS_PTR
	movwf	FSR
	
	call	PointToNextProcessBufferByte	; Point MIDI_MSG_PROCESS_PTR to velocity
	
	; Check MIDI channel
	bsf	STATUS,RP1	; Bank 2 Select
	movfw	vBASE_CHA
	bcf	STATUS,RP1	; Bank 0 Select
	subwf	MTEMP1,w
	rnz
	
	movfw	INDF		; Get and save note value
	movwf	MTEMP1
	
ProcessMIDI_NoteOff_1
	bsf	STATUS,RP0	; Bank 1 Select
	
	; Check if there are unreleased notes on note stack
	movfw	NOTE_STACK_PTR
	addlw	0xF0		; Test if >= 0x10
	jc	ProcessMIDI_NoteOff_LastNote
	
	; Search for note on stack (note in MTEMP1)
	
	decf	NOTE_STACK_PTR,w
	movwf	MTEMP2		; Counter
	
	bsf	STATUS,IRP	; Indirectly address banks 2-3
	addlw	low NOTE_STACK	; Point to top of stack
	movwf	FSR
	
ProcessMIDI_NoteOff_Loop
	movfw	INDF
	subwf	MTEMP1,w
	jz	ProcessMIDI_NoteOff_FoundNote
	
	incf	FSR
	incf	MTEMP2		; Loop until at bottom of stack
	btfss	MTEMP2,4
	goto	ProcessMIDI_NoteOff_Loop
	
	; Note not found on stack
	; Read next CV value off stack top and change CV
	movfw	NOTE_STACK_PTR
	addlw	low (NOTE_STACK - 1)
	movwf	FSR
	movfw	INDF
	
	goto	ProcessMIDI_NoteOff_ChangeCV
	
	
ProcessMIDI_NoteOff_FoundNote
	incf	MTEMP2,w	; If MTEMP2 + 1 = NOTE_STACK_PTR, skip shift
	subwf	NOTE_STACK_PTR,w
	jz	ProcessMIDI_NoteOff_FoundNote_2

ProcessMIDI_NoteOff_FoundNote_1
	; Remove note and shift stack
	decf	FSR
	movfw	INDF
	incf	FSR
	movwf	INDF
	decf	FSR
	
	decf	MTEMP2		; Loop until at top of stack
	incf	MTEMP2,w
	subwf	NOTE_STACK_PTR,w
	jnz	ProcessMIDI_NoteOff_FoundNote_1

ProcessMIDI_NoteOff_FoundNote_2
	incf	FSR
	movfw	INDF		; Read next CV value off stack top
	
	; Note value in w
	; RAM bank 1
	bcf	MFLAGS,NOTE_ON
	call	NoteToCV
	
	incf	NOTE_STACK_PTR	; Adjust stack pointer
	goto	ProcessMIDI_NoteOff_End
	
	
ProcessMIDI_NoteOff_ChangeCV
	; Note value in w
	; RAM bank 1
	bcf	MFLAGS,NOTE_ON
	call	NoteToCV
	goto	ProcessMIDI_NoteOff_End
	
	
ProcessMIDI_NoteOff_LastNote
	movlw	0x11		; Reset note stack pointer
	movwf	NOTE_STACK_PTR
	
	bcf	STATUS,RP0	; Bank 0 Select
	bsf	PORTE,GATE1	; Gate low
	bcf	PORTE,GATE3	; Inverted gate high
	
	bsf	PCLATH,3	; Select program bank 3
	movlw	3
	call	CheckAndTriggerEnvAttacks
	movlw	1
	call	CheckAndTriggerEnvReleases
	bcf	PCLATH,3	; Select program bank 2
	
	; Gate on note off
	
	bsf	STATUS,RP1	; Bank 2 Select
	
	movfw	vN_O_GATE	; Save LSB in MTEMP3 for now
	movwf	MTEMP3
	movfw	vN_O_GATE_H
	
	bsf	STATUS,RP0	; Bank 1 Select
	bcf	STATUS,RP1
	
	movwf	AARGB0		; Put MSB in AARGB0
	
	iorwf	MTEMP3,w	; If both LSB+MSB 0 then no gate on note off
	jz	ProcessMIDI_NoteOff_End
	
	movfw	MTEMP3		; Put LSB in AARGB1
	movwf	AARGB1
	
	; * 13
	movlw	D'13'
	movwf	BARGB0
	call	FXM1507U
	
	; Put result in NOTE_OFF_GATE_STEPS_L & NOTE_OFF_GATE_STEPS_H
	movfw	AARGB2
	movwf	NOTE_OFF_GATE_STEPS_L
	movfw	AARGB1
	movwf	NOTE_OFF_GATE_STEPS_H
	
	; Set note off gate high
	bcf	STATUS,RP0	; Bank 0 Select
	bcf	PORTE,GATE2	; Note off gate high
	
	bsf	PCLATH,3	; Select program bank 3
	movlw	2
	call	CheckAndTriggerEnvAttacks
	bcf	PCLATH,3	; Select program bank 2
	
ProcessMIDI_NoteOff_End
	bcf	STATUS,RP0	; Bank 0 Select
	return
	
	
;***********************************************************************
	
; RAM Bank 0
ProcessMIDI_NoteOn
	call	PointToNextProcessBufferByte	; Point FSR to note value
	movfw	MIDI_MSG_PROCESS_PTR
	movwf	FSR
	
	call	PointToNextProcessBufferByte	; Point MIDI_MSG_PROCESS_PTR to velocity
	
	; Check MIDI channel
	bsf	STATUS,RP1	; Bank 2 Select
	movfw	vBASE_CHA
	bcf	STATUS,RP1	; Bank 0 Select
	subwf	MTEMP1,w
	rnz
	
	movfw	INDF		; Get and save note value
	movwf	MTEMP1
	
	movfw	MIDI_MSG_PROCESS_PTR	; Read velocity
	movwf	FSR
	movfw	INDF
	jz	ProcessMIDI_NoteOff_1		; If velocity=0, treat as note-off
	
	bsf	STATUS,RP0	; Bank 1 Select
	
	; Velocity value in w
	; RAM bank 1
	call	VelocityToCV
	
	bsf	STATUS,IRP	; Indirectly address banks 2-3
	decf	NOTE_STACK_PTR,w	; Load stack pointer - 1 into w
	jz	ProcessMIDI_NoteOn_1	; If w=0, no stack left so don't push note
	
	decf	NOTE_STACK_PTR		; Decrement stack pointer
	
	movwf	FSR			; Point to stack byte
	movlw	low (NOTE_STACK - 1)
	addwf	FSR,f
	
	movfw	MTEMP1			; Push note value onto stack
	movwf	INDF
	
ProcessMIDI_NoteOn_1
	movfw	MTEMP1		; Get note
	
	; Note value in w
	; RAM bank 1
	bsf	MFLAGS,NOTE_ON
	call	NoteToCV
	
	bcf	STATUS,RP0	; Bank 0 Select
	
	btfsc	PORTE,GATE1	; If gates already off, no need to re-trigger
	goto	ProcessMIDI_NoteOn_GatesOn
	
	bsf	STATUS,RP1	; Bank 2 Select
	
	movfw	vL_RETRIG	; If legato retrigger mode is off, no need to re-trigger
	bcf	STATUS,RP1	; Bank 0 Select
	jz	ProcessMIDI_NoteOn_GatesOn
	
	; Turn off gates momentarily before turning on again
	bsf	PORTE,GATE1	; Gate low
	bcf	PORTE,GATE3	; Inverted gate high
	wait_us_M 20	; Does work just to wait 10us, but wait 20us to be safe

ProcessMIDI_NoteOn_GatesOn
	btfss	PORTE,GATE1	; If gates already on, don't re-trigger envelopers
	goto	ProcessMIDI_NoteOn_GatesOn_1
	
	bsf	PCLATH,3	; Select program bank 3
	movlw	1
	call	CheckAndTriggerEnvAttacks
	movlw	3
	call	CheckAndTriggerEnvReleases
	bcf	PCLATH,3	; Select program bank 2
	
ProcessMIDI_NoteOn_GatesOn_1
	bcf	PORTE,GATE1	; Gate high
	bsf	PORTE,GATE3	; Inverted gate low
	
	return
	
	
; Velocity value in w
; RAM bank 1
VelocityToCV
	; Velocity -> DAC value
	; 0x1FFE - (66 * Velocity * Amount) >> 8
	
	; Velocity * Amount
	movwf	AARGB0
	bcf	STATUS,RP0	; Bank 2 Select
	bsf	STATUS,RP1
	movfw	vVELOCITY
	bsf	STATUS,RP0	; Bank 1 Select
	bcf	STATUS,RP1
	rz			; Return if Amount=0 (disabled)
	movwf	BARGB0
	call	FXM0808U
	
	; (* 66) >> 8 -> AARGB0,AARGB1
	movlw	D'66'
	movwf	BARGB0
	call	FXM1507U
	
	; 0x1FFE - AARGB0,AARGB1 -> W,MTEMP2
	movfw	AARGB1
	sublw	0xFE
	movwf	MTEMP2
	movfw	AARGB0
	sc
	incfsz	AARGB0,w
	sublw	0x1F
	
	; Write to CV4
	bcf	INTCON,GIE		; Disable interrupts while changing CV RAM
	movwf	CV4_VALUE_H
	movfw	MTEMP2
	movwf	CV4_VALUE_L
	bsf	INTCON,GIE		; Enable interrupts again	
	
	return
	
	
; Note value in w
; RAM bank 1
NoteToCV
	; Note -> DAC value
	; (127 - N) * 0x5B + 0x03E9
	
	; (127 - N)
	sublw	D'127'
	; * 0x5B
	movwf	AARGB0
	movlw	0x5B
	movwf	BARGB0
	call	FXM0707U
	
	; + 0x03E9
	movlw	0xE9
	addwf	AARGB1,w
	movwf	TMP_NOTE_VALUE
	movlw	0x03
	btfsc	STATUS,C
	addlw	1
	addwf	AARGB0,w
	movwf	TMP_NOTE_VALUE_H
	
	; Base DAC value in TMP_NOTE_VALUE
	
	
	; Add in octave shift
	movlw	low 0xFBBC ; -0x444
	movwf	AARGB1
	movlw	high 0xFBBC ; -0x444
	movwf	AARGB0
	bcf	STATUS,RP0	; Bank 2 Select
	bsf	STATUS,RP1
	movfw	vOCTAVE_1
	bsf	STATUS,RP0	; Bank 1 Select
	bcf	STATUS,RP1
	movwf	BARGB0
	call	FXM1608S
	
	
	bsf	STATUS,RP1	; Bank 3 Select
	clrf	PORT1_STEPS_L	; Set steps to 0 to stop any active portamento
	clrf	PORT1_STEPS_H
	
	; Portamento on?
	bcf	STATUS,RP0	; Bank 2 Select
	movfw	vPORTA
	bsf	STATUS,RP0	; Bank 1 Select
	bcf	STATUS,RP1
	jnz	NoteToCV_CV1_Portamento
	
	
NoteToCV_1
	AddAndCheck14bit  TMP_NOTE_VALUE, AARGB2, NOTE1_AMT_L

	; Add in pitch bend and set CV
	bcf	INTCON,GIE		; Disable interrupts while changing CV RAM
	AddAndCheck14bit  NOTE1_AMT_L, BEND1_AMT_L, CV1_VALUE_L
	bsf	INTCON,GIE		; Enable interrupts again
	
	
	bcf	STATUS,RP0	; Bank 2 Select
	bsf	STATUS,RP1

NoteToCV_CheckCV2	
	; Check portamento 2
	movfw	vPORTA_T2
	iorwf	vPORTA_T2_H,w
	jnz	NoteToCV_CV2
	
	; Check pitch bend 2
	movfw	vP_BEND_2
	jnz	NoteToCV_CV2
	
	; Check octave 2
	movfw	vOCTAVE_2
	jz	NoteToCV_CheckCV3
	
	
NoteToCV_CV2
	bsf	STATUS,RP0	; Bank 1 Select
	bcf	STATUS,RP1
	
	; Add in octave shift
	movlw	low 0xFBBC ; -0x444
	movwf	AARGB1
	movlw	high 0xFBBC ; -0x444
	movwf	AARGB0
	bcf	STATUS,RP0	; Bank 2 Select
	bsf	STATUS,RP1
	movfw	vOCTAVE_2
	bsf	STATUS,RP0	; Bank 1 Select
	bcf	STATUS,RP1
	movwf	BARGB0
	call	FXM1608S
	
	bsf	STATUS,RP1	; Bank 3 Select
	clrf	PORT2_STEPS_L	; Set steps to 0 to stop any active portamento
	clrf	PORT2_STEPS_H
	
	; Portamento?
	bcf	STATUS,RP0	; Bank 2 Select
	movfw	vPORTA_T2
	iorwf	vPORTA_T2_H,w
	jz	NoteToCV_CV2_1
	
	movfw	vPORTA
	jnz	NoteToCV_CV2_Portamento
	
	
NoteToCV_CV2_1
	bsf	STATUS,RP0	; Bank 1 Select
	bcf	STATUS,RP1
NoteToCV_CV2_2
	AddAndCheck14bit  TMP_NOTE_VALUE, AARGB2, NOTE2_AMT_L
	
	
	; Add in pitch bend and set CV
	bcf	INTCON,GIE		; Disable interrupts while changing CV RAM
	AddAndCheck14bit  NOTE2_AMT_L, BEND2_AMT_L, CV2_VALUE_L
	bsf	INTCON,GIE		; Enable interrupts again
	
	
	bcf	STATUS,RP0	; Bank 2 Select
	bsf	STATUS,RP1
	
NoteToCV_CheckCV3
	; Check portamento 3
	movfw	vPORTA_T3
	iorwf	vPORTA_T3_H,w
	jnz	NoteToCV_CV3
	
	; Check pitch bend 3
	movfw	vP_BEND_3
	jnz	NoteToCV_CV3
	
	; Check octave 3
	movfw	vOCTAVE_3

	bsf	STATUS,RP0	; Bank 1 Select
	bcf	STATUS,RP1

	rz			; Return if octave 3 = 0


NoteToCV_CV3
	bsf	STATUS,RP0	; Bank 1 Select
	bcf	STATUS,RP1
	
	; Add in octave shift
	movlw	low 0xFBBC ; -0x444
	movwf	AARGB1
	movlw	high 0xFBBC ; -0x444
	movwf	AARGB0
	bcf	STATUS,RP0	; Bank 2 Select
	bsf	STATUS,RP1
	movfw	vOCTAVE_3
	bsf	STATUS,RP0	; Bank 1 Select
	bcf	STATUS,RP1
	movwf	BARGB0
	call	FXM1608S
	
	bsf	STATUS,RP1	; Bank 3 Select
	clrf	PORT3_STEPS_L	; Set steps to 0 to stop any active portamento
	clrf	PORT3_STEPS_H
	
	; Portamento?
	bcf	STATUS,RP0	; Bank 2 Select
	movfw	vPORTA_T3
	iorwf	vPORTA_T3_H,w
	jz	NoteToCV_CV3_1		; vPORTA_T == 0 ?
	
	movfw	vPORTA
	jnz	NoteToCV_CV3_Portamento	; vPORTA != 0 ?
	
	
NoteToCV_CV3_1
	bsf	STATUS,RP0	; Bank 1 Select
	bcf	STATUS,RP1
NoteToCV_CV3_2
	AddAndCheck14bit  TMP_NOTE_VALUE, AARGB2, NOTE3_AMT_L
	
	
	; Add in pitch bend and set CV
	bcf	INTCON,GIE		; Disable interrupts while changing CV RAM
	AddAndCheck14bit  NOTE3_AMT_L, BEND3_AMT_L, CV3_VALUE_L
	bsf	INTCON,GIE		; Enable interrupts again
	
	return
	


; Start in RAM Bank 1, ends in RAM Bank 2
NoteToCV_CV_Portamento	macro	vPORTA_T, PORT_STEPS, PORT_STEP_AMT, PORT_REM_AMT, NOTE_AMT, P_NEG

	; Save new note value in MTEMP3+MTEMP4
	AddAndCheck14bit  TMP_NOTE_VALUE, AARGB2, MTEMP3
	

	; STEPS = PORT_TIME * 13
	
	bcf	STATUS,RP0	; Bank 2 Select
	bsf	STATUS,RP1
	; Get portamento time
	movfw	vPORTA_T
	movwf	MTEMP1
	movfw	(vPORTA_T+1)
	bsf	STATUS,RP0	; Bank 3 Select
	movwf	AARGB0
	movfw	MTEMP1
	movwf	AARGB1
	
	; * 13
	movlw	D'13'
	movwf	BARGB0
	call	FXM1507U
	
	; STEPS -> BARG & MTEMP1+MTEMP2
	movfw	AARGB2
	movwf	BARGB1
	movwf	MTEMP1
	movfw	AARGB1
	movwf	BARGB0
	movwf	MTEMP2
	
	; DAC_CHANGE = New value - Current value
	clrf	AARGB2		; Load new DAC value into AARG
	movfw	MTEMP3
	movwf	AARGB1
	movfw	MTEMP4
	movwf	AARGB0

	; Copy current note amount value to MTEMP3+MTEMP4+MTEMP5
	bcf	STATUS,RP1	; Bank 1 Select
	movfw	NOTE_AMT
	bsf	STATUS,RP1	; Bank 3 Select
	movwf	MTEMP3
	bcf	STATUS,RP1	; Bank 1 Select
	movfw	(NOTE_AMT+1)
	bsf	STATUS,RP1	; Bank 3 Select
	movwf	MTEMP4
	bcf	STATUS,RP1	; Bank 1 Select
	movfw	(NOTE_AMT+2)
	bsf	STATUS,RP1	; Bank 3 Select
	movwf	MTEMP5

	; AARG - NOTE_AMT => AARG
	SUB2424 AARGB2, MTEMP3
	
	; If negative, take abs value and set negative flag
	bcf	PORT_FLAGS,P_NEG	; Clear flag
	btfss	AARGB0,7
	goto	$+D'15'			; Jump ahead to just after TwosComp24 macro
	bsf	PORT_FLAGS,P_NEG	; Set flag
	TwosComp24  AARGB2		; 0 - AARG (13 instructions)
	
	
	; STEP_AMT = DAC_CHANGE * 0x100 / STEPS
	;                           Remainder -> REM_AMT
	call	FXD2416U
	
	movfw	AARGB2
	movwf	PORT_STEP_AMT
	movfw	AARGB1
	movwf	(PORT_STEP_AMT+1)
	movfw	REMB1
	movwf	PORT_REM_AMT
	movfw	REMB0
	movwf	(PORT_REM_AMT+1)

	bcf	INTCON,GIE	; Disable interrupts while changing steps
	movfw	MTEMP1
	movwf	PORT_STEPS
	movfw	MTEMP2
	movwf	(PORT_STEPS+1)	
	bsf	INTCON,GIE	; Enable interrupts again
	
	bcf	STATUS,RP0	; Bank 2 Select
endm


; RAM Bank 1, ends in RAM Bank 2
; vPORTA in w
NoteToCV_CV1_Portamento
	movwf	MTEMP3		; Save vPORTA
	
	movlw	0x10			; Check if any other keys are down
	subwf	NOTE_STACK_PTR,w
	jnz	NoteToCV_CV1_Portamento_Legato
	
	; Not played legato
	decf	MTEMP3		; If Legato 1 mode, jump back
	jz	NoteToCV_1
	decf	MTEMP3		; If Legato 2 mode, jump back
	jz	NoteToCV_1
	decf	MTEMP3		; If Legato 3 mode, jump back
	jz	NoteToCV_1
	goto	NoteToCV_CV1_Portamento_End	; Else: Portamento always on
	
	; Played legato
NoteToCV_CV1_Portamento_Legato
	decf	MTEMP3		; If Legato 1 mode and note-on, jump back
	jnz	$+3
	btfsc	MFLAGS,NOTE_ON
	goto	NoteToCV_1
	
	decf	MTEMP3		; If Legato 2 mode and note-off, jump back
	jnz	$+3
	btfss	MFLAGS,NOTE_ON
	goto	NoteToCV_1
	
NoteToCV_CV1_Portamento_End
	NoteToCV_CV_Portamento	vPORTA_T1, PORT1_STEPS_L, PORT1_STEP_AMT_LL, PORT1_REM_AMT_LL, NOTE1_AMT_LL, P1_NEG
	goto	NoteToCV_CheckCV2
	
; RAM Bank 2
; vPORTA in w
NoteToCV_CV2_Portamento
	bsf	STATUS,RP0	; Bank 1 Select
	bcf	STATUS,RP1
	
	movwf	MTEMP3		; Save vPORTA
	
	movlw	0x10			; Check if any other keys are down
	subwf	NOTE_STACK_PTR,w
	jnz	NoteToCV_CV2_Portamento_Legato
	
	; Not played legato
	decf	MTEMP3		; If Legato 1 mode, jump back
	jz	NoteToCV_CV2_2
	decf	MTEMP3		; If Legato 2 mode, jump back
	jz	NoteToCV_CV2_2
	decf	MTEMP3		; If Legato 3 mode, jump back
	jz	NoteToCV_CV2_2
	goto	NoteToCV_CV2_Portamento_End	; Else: Portamento always on
	
	; Played legato
NoteToCV_CV2_Portamento_Legato
	decf	MTEMP3		; If Legato 1 mode and note-on, jump back
	jnz	$+3
	btfsc	MFLAGS,NOTE_ON
	goto	NoteToCV_CV2_2
	
	decf	MTEMP3		; If Legato 2 mode and note-off, jump back
	jnz	$+3
	btfss	MFLAGS,NOTE_ON
	goto	NoteToCV_CV2_2
	
NoteToCV_CV2_Portamento_End
	NoteToCV_CV_Portamento	vPORTA_T2, PORT2_STEPS_L, PORT2_STEP_AMT_LL, PORT2_REM_AMT_LL, NOTE2_AMT_LL, P2_NEG
	goto	NoteToCV_CheckCV3
	
; RAM Bank 2, ends in RAM Bank 1
; vPORTA in w
NoteToCV_CV3_Portamento
	bsf	STATUS,RP0	; Bank 1 Select
	bcf	STATUS,RP1
	
	movwf	MTEMP3		; Save vPORTA
	
	movlw	0x10			; Check if any other keys are down
	subwf	NOTE_STACK_PTR,w
	jnz	NoteToCV_CV3_Portamento_Legato
	
	; Not played legato
	decf	MTEMP3		; If Legato 1 mode, jump back
	jz	NoteToCV_CV3_2
	decf	MTEMP3		; If Legato 2 mode, jump back
	jz	NoteToCV_CV3_2
	decf	MTEMP3		; If Legato 3 mode, jump back
	jz	NoteToCV_CV3_2
	goto	NoteToCV_CV3_Portamento_End	; Else: Portamento always on
	
	; Played legato
NoteToCV_CV3_Portamento_Legato
	decf	MTEMP3		; If Legato 1 mode and note-on, jump back
	jnz	$+3
	btfsc	MFLAGS,NOTE_ON
	goto	NoteToCV_CV3_2
	
	decf	MTEMP3		; If Legato 2 mode and note-off, jump back
	jnz	$+3
	btfss	MFLAGS,NOTE_ON
	goto	NoteToCV_CV3_2
	
NoteToCV_CV3_Portamento_End
	NoteToCV_CV_Portamento	vPORTA_T3, PORT3_STEPS_L, PORT3_STEP_AMT_LL, PORT3_REM_AMT_LL, NOTE3_AMT_LL, P3_NEG
	
	bsf	STATUS,RP0	; Bank 1 Select
	bcf	STATUS,RP1
	return
	
	
;***********************************************************************
	
; RAM Bank 0
ProcessMIDI_PolyAftertouch
	call	PointToNextProcessBufferByte
	call	PointToNextProcessBufferByte
	return
	
	
;***********************************************************************
	
; RAM Bank 0
ProcessMIDI_ControlChange
	call	PointToNextProcessBufferByte	; Get and save controller number
	movfw	MIDI_MSG_PROCESS_PTR
	movwf	FSR
	movfw	INDF
	movwf	MTEMP2
	call	PointToNextProcessBufferByte	; Get and save controller value
	movfw	MIDI_MSG_PROCESS_PTR
	movwf	FSR
	movfw	INDF
	movwf	MTEMP3
	
	; Check MIDI channel against control channel
	bsf	STATUS,RP1	; Bank 2 Select
	decf	vCTRL_CHA,w
	bcf	STATUS,RP1	; Bank 0 Select
	subwf	MTEMP1,w
	bsf	PCLATH,3	; Select program bank 3
	snz
	call	ChangeParamViaMIDI
	bcf	PCLATH,3	; Select program bank 2
	
	; Check MIDI channel
	bsf	STATUS,RP1	; Bank 2 Select
	movfw	vBASE_CHA
	bcf	STATUS,RP1	; Bank 0 Select
	subwf	MTEMP1,w
	rnz
	
	; Scan sources for this CC or for CC learn
	bsf	STATUS,IRP	; Indirectly address banks 2-3
	movlw	low vSOURCE_CV1
	movwf	FSR
	movlw	D'15'
	movwf	MTEMP1
	
ProcessMIDI_ControlChange_Loop
	movlw	3		; Subtract 3 => 0=CC Learn, 1-128=CC
	subwf	INDF,w
	movwf	MTEMP4
	snz
	call	ProcessMIDI_ControlChange_CCLearn	; Learn CC (will be written to INDF and MTEMP4)
	
	decf	MTEMP4		; Decrement to give actual CC value
	
	movfw	MTEMP2		; Check if matches CC number
	subwf	MTEMP4,w
	snz
	call	Controller2CV
	
ProcessMIDI_ControlChange_Loop_1
	incf	FSR
	decfsz	MTEMP1
	goto	ProcessMIDI_ControlChange_Loop
	
	return
	

ProcessMIDI_ControlChange_CCLearn
	movfw	MTEMP2		; Write CC to MTEMP4
	movwf	MTEMP4
	addlw	4		; Add 4 to write to INDF
	movwf	INDF
	
	bcf	PCLATH,4	; Select program bank 0
	movfw	FSR		; Save FSR
	movwf	MTEMP5
	bcf	INTCON,GIE	; Disable interrupts while updating display
	call	UpdateDisplay_ValueOnly
	bsf	INTCON,GIE	; Enable interrupts again
	bsf	PCLATH,4	; Select program bank 2
	bsf	STATUS,IRP	; Restore IRP
	movfw	MTEMP5		; Restore FSR
	movwf	FSR
	return
	

; Status:
; RAM Bank 0
; IRP addressing banks 2-3
;
; Input:
; FSR = vSOURCE_CVn
; MTEMP1 = 16 - CV number
; MTEMP3 = Controller value
;
; Changes:
; MTEMP5
; Maths variables in bank 1
Controller2CV
	bsf	STATUS,RP0	; RAM Bank 1 Select (to call maths routines)
	
	movfw	FSR		; Save FSR
	movwf	MTEMP5
	
	movlw	D'15'		; Copy amount to AARGB0
	addwf	FSR,f
	movfw	INDF
	movwf	AARGB0
	
	movfw	MTEMP3		; Copy controller value to BARGB0
	movwf	BARGB0
	movlw	0x80		; If a pitch-bend value of -128 received, treat as -127
	subwf	BARGB0,w
	snz
	incf	BARGB0
	
	bcf	STATUS,IRP	; Set IRP (Indirectly address banks 0-1)
	
	movlw	low (CV15_VALUE_H + 1)	; Point to CVn_VALUE_L
	movwf	FSR
	movfw	MTEMP1
	subwf	FSR
	subwf	FSR
	
	; DAC value ~= 0x1FFE - (132 * Value * Amount) >> 8
	
	; Value * Amount => AARGB0,AARGB1
	call	FXM0808S
	
	
	; (AARGB0,AARGB1 * 132) >> 8 => AARGB0,AARGB1
	; (AARGB0,AARGB1 * 128 + AARGB0,AARGB1 * 4) >> 8 => AARGB0,AARGB1
	
	; AARGB0,AARGB1 * 128 => SPARE0,SPARE1,BARGB1
	movfw	AARGB0
	movwf	SPARE0
	movfw	AARGB1
	movwf	SPARE1
	clrf	BARGB1
	bcf	STATUS,C
	rrf	SPARE0
	rrf	SPARE1
	rrf	BARGB1
	btfsc	SPARE0,6	; Copy highest-but-1 bit (in case of -ve AARG)
	bsf	SPARE0,7
	
	; AARGB0,AARGB1 * 4 => AARGB0,AARGB1,AARGB2
	movlw	D'4'
	movwf	BARGB0
	call	FXM1608S
	
	; (AARGB0,AARGB1,AARGB2 + SPARE0,SPARE1,BARGB1) >> 8 => AARGB0,AARGB1
	movfw	BARGB1
	addwf	AARGB2,f
	movfw	SPARE1
	snc
	incfsz	SPARE1,w
	addwf	AARGB1,f
	movfw	SPARE0
	snc
	incfsz	SPARE0,w
	addwf	AARGB0,f
	
	
	; 0x1FFE - AARGB0,AARGB1 -> W,CVn_VALUE_L
	movfw	AARGB1
	sublw	0xFE
	bcf	INTCON,GIE	; Disable interrupts while changing CV RAM
	movwf	INDF
	movfw	AARGB0
	sc
	incfsz	AARGB0,w
	sublw	0x1F
	
	; W -> CVn_VALUE_H
	incf	FSR,f
	movwf	INDF
	bsf	INTCON,GIE	; Enable interrupts again
	
	
	bcf	STATUS,RP0	; RAM Bank 0 Select
	
	bsf	STATUS,IRP	; Restore IRP (Indirectly address banks 2-3)
	movfw	MTEMP5		; Restore FSR
	movwf	FSR
	return
	
	
;***********************************************************************
	
; RAM Bank 0
ProcessMIDI_ProgramChange
	call	PointToNextProcessBufferByte
	return
	
	
;***********************************************************************
	
; RAM Bank 0
ProcessMIDI_Aftertouch
	call	PointToNextProcessBufferByte	; Get and save controller value
	movfw	MIDI_MSG_PROCESS_PTR
	movwf	FSR
	movfw	INDF
	movwf	MTEMP3
	
	; Check MIDI channel
	bsf	STATUS,RP1	; Bank 2 Select
	movfw	vBASE_CHA
	bcf	STATUS,RP1	; Bank 0 Select
	subwf	MTEMP1,w
	rnz
	
	; Scan sources for aftertouch
	bsf	STATUS,IRP	; Indirectly address banks 2-3
	movlw	low vSOURCE_CV1
	movwf	FSR
	movlw	D'15'
	movwf	MTEMP1
	
ProcessMIDI_Aftertouch_Loop
	movlw	2		; Check for Aftertouch (2)
	subwf	INDF,w
	snz
	call	Controller2CV
	
	incf	FSR
	decfsz	MTEMP1
	goto	ProcessMIDI_Aftertouch_Loop
	
	return
	
	
;***********************************************************************
	
; RAM Bank 0
ProcessMIDI_PitchBend
	
	; Check MIDI channel
	bsf	STATUS,RP1	; Bank 2 Select
	movfw	vBASE_CHA
	bcf	STATUS,RP1	; Bank 0 Select
	subwf	MTEMP1,w
	jnz	ProcessMIDI_PitchBend_End2
	
	
	; Get 14 bit pitch bend value
	
	call	PointToNextProcessBufferByte
	movfw	MIDI_MSG_PROCESS_PTR
	movwf	FSR
	rlf	INDF,w		; Get LSB and rotate left 1 bit, 0 -> C
	movwf	MTEMP1
	
	call	PointToNextProcessBufferByte
	movfw	MIDI_MSG_PROCESS_PTR
	movwf	FSR
	rrf	INDF		; Get MSB and rotate right 1 bit, LSbit -> C
	
	rrf	MTEMP1		; Rotate LSB right 1 bit and pick up C bit
	
	bsf	STATUS,RP1	; Bank 2 Select
	decf	vP_BEND_1,w	; Get pitch bend amount
	
	bsf	STATUS,RP0	; Bank 1 Select
	bcf	STATUS,RP1
	
	; PBA in w, PB in LSB:MTEMP1 MSB:INDF
	; Bend amount returned in TMP_NOTE_VALUE_H,TMP_NOTE_VALUE
	call	PitchBendToDacOffset
	
	; TMP_NOTE_VALUE_H,TMP_NOTE_VALUE => BEND1_AMT_H,BEND1_AMT_L
	movfw	TMP_NOTE_VALUE_H
	movwf	BEND1_AMT_H
	movfw	TMP_NOTE_VALUE
	movwf	BEND1_AMT_L
	
	bcf	INTCON,GIE		; Disable interrupts while changing CV RAM
	AddAndCheck14bit  NOTE1_AMT_L, BEND1_AMT_L, CV1_VALUE_L
	bsf	INTCON,GIE		; Enable interrupts again
	
	
; Check CV2
	bcf	STATUS,RP0	; Bank 2 Select
	bsf	STATUS,RP1
	
	; Check portamento 2
	movfw	vPORTA_T2
	iorwf	vPORTA_T2_H,w
	jnz	ProcessMIDI_PitchBend_CV2
	
	; Check pitch bend 2
	movfw	vP_BEND_2
	jnz	ProcessMIDI_PitchBend_CV2
	
	; Check octave 2
	movfw	vOCTAVE_2
	jz	ProcessMIDI_PitchBend_CheckCV3
	
	
ProcessMIDI_PitchBend_CV2
	decf	vP_BEND_2,w	; Get pitch bend amount
	
	bsf	STATUS,RP0	; Bank 1 Select
	bcf	STATUS,RP1
	
	; PBA in w, PB in LSB:MTEMP1 MSB:INDF
	; Bend amount returned in TMP_NOTE_VALUE_H,TMP_NOTE_VALUE
	call	PitchBendToDacOffset
	
	; TMP_NOTE_VALUE_H,TMP_NOTE_VALUE => BEND2_AMT_H,BEND2_AMT_L
	movfw	TMP_NOTE_VALUE_H
	movwf	BEND2_AMT_H
	movfw	TMP_NOTE_VALUE
	movwf	BEND2_AMT_L
	
	bcf	INTCON,GIE		; Disable interrupts while changing CV RAM
	AddAndCheck14bit  NOTE2_AMT_L, BEND2_AMT_L, CV2_VALUE_L
	bsf	INTCON,GIE		; Enable interrupts again
	
	
	bcf	STATUS,RP0	; Bank 2 Select
	bsf	STATUS,RP1
	
ProcessMIDI_PitchBend_CheckCV3
	; Check portamento 3
	movfw	vPORTA_T3
	iorwf	vPORTA_T3_H,w
	jnz	ProcessMIDI_PitchBend_CV3
	
	; Check pitch bend 3
	movfw	vP_BEND_3
	jnz	ProcessMIDI_PitchBend_CV3
	
	; Check octave 3
	movfw	vOCTAVE_3
	jz	ProcessMIDI_PitchBend_End


ProcessMIDI_PitchBend_CV3
	decf	vP_BEND_3,w	; Get pitch bend amount
	
	bsf	STATUS,RP0	; Bank 1 Select
	bcf	STATUS,RP1
	
	; PBA in w, PB in LSB:MTEMP1 MSB:INDF
	; Bend amount returned in TMP_NOTE_VALUE_H,TMP_NOTE_VALUE
	call	PitchBendToDacOffset
	
	; TMP_NOTE_VALUE_H,TMP_NOTE_VALUE => BEND3_AMT_H,BEND3_AMT_L
	movfw	TMP_NOTE_VALUE_H
	movwf	BEND3_AMT_H
	movfw	TMP_NOTE_VALUE
	movwf	BEND3_AMT_L
	
	bcf	INTCON,GIE		; Disable interrupts while changing CV RAM
	AddAndCheck14bit  NOTE3_AMT_L, BEND3_AMT_L, CV3_VALUE_L
	bsf	INTCON,GIE		; Enable interrupts again
	
	
ProcessMIDI_PitchBend_End
	clrf	STATUS		; Bank 0 Select
	
	; Now to check if pitch bend needs to be sent to any other CVs
	
	rlf	MTEMP1		; Get most significant 8 bits into MTEMP3
	rlf	INDF
	rlf	MTEMP1
	rlf	INDF,w
	movwf	MTEMP3
	
	movlw	D'128'		; Subtract 128 to give range of -128 to 127
	subwf	MTEMP3,f
	
	; Scan sources for pitch bend
	bsf	STATUS,IRP	; Indirectly address banks 2-3
	movlw	low vSOURCE_CV1
	movwf	FSR
	movlw	D'15'
	movwf	MTEMP1
	
ProcessMIDI_PitchBend_Loop
	decf	INDF,w		; Check for pitch bend (1)
	snz
	call	Controller2CV
	
	incf	FSR
	decfsz	MTEMP1
	goto	ProcessMIDI_PitchBend_Loop
	
	return
	
	
ProcessMIDI_PitchBend_End2
	call	PointToNextProcessBufferByte
	call	PointToNextProcessBufferByte
	return
	
	
;***********************************************************************
	
; RAM Bank 0
ProcessMIDI_System
	return
	
	
;***********************************************************************
;***********************************************************************
	
	

; Maths Functions
#include <maths.inc>


;#######################################################################


	; Program bank 3
	
	org	0x1800
	
	nop	; To be sure to get a compiler warning if stuff in bank 2 overspills

; String table
#include <strings.inc>

	
;***********************************************************************
;***********************************************************************


; RAM Bank 0
; Gate number in w
CheckAndTriggerEnvAttacks
	; Trigger attack phase of envelopes?
	bsf	STATUS,RP1	; Bank 2 Select
	movwf	ETEMP1
	subwf	vENV1_TRG,w
	jnz	$+7
	bsf	STATUS,RP0	; Bank 3 Select
	bcf	ENV_FLAGS,ENV1_PHASEA		; Attack phase
	bcf	STATUS,RP0	; Bank 0 Select
	bcf	STATUS,RP1
	bsf	MENU_FLAGS,ENV1_CALC_PENDING	; Trigger ENV1 calculation
	bsf	STATUS,RP1	; Bank 2 Select
	movfw	ETEMP1
	subwf	vENV2_TRG,w
	jnz	$+6
	bsf	STATUS,RP0	; Bank 3 Select
	bcf	ENV_FLAGS,ENV2_PHASEA		; Attack phase
	bcf	STATUS,RP0	; Bank 0 Select
	bcf	STATUS,RP1
	bsf	MENU_FLAGS,ENV2_CALC_PENDING	; Trigger ENV2 calculation
	bcf	STATUS,RP1	; Bank 0 Select
	return
	
; RAM Bank 0
; Gate number in w
CheckAndTriggerEnvReleases
	; Trigger release phase of envelopes?
	bsf	STATUS,RP1	; Bank 2 Select
	movwf	ETEMP1
	subwf	vENV1_TRG,w
	jnz	$+8
	bsf	STATUS,RP0	; Bank 3 Select
	bsf	ENV_FLAGS,ENV1_PHASEA		; Release phase
	bsf	ENV_FLAGS,ENV1_PHASEB
	bcf	STATUS,RP0	; Bank 0 Select
	bcf	STATUS,RP1
	bsf	MENU_FLAGS,ENV1_CALC_PENDING	; Trigger ENV1 calculation
	bsf	STATUS,RP1	; Bank 2 Select
	movfw	ETEMP1
	subwf	vENV2_TRG,w
	jnz	$+7
	bsf	STATUS,RP0	; Bank 3 Select
	bsf	ENV_FLAGS,ENV2_PHASEA		; Release phase
	bsf	ENV_FLAGS,ENV2_PHASEB
	bcf	STATUS,RP0	; Bank 0 Select
	bcf	STATUS,RP1
	bsf	MENU_FLAGS,ENV2_CALC_PENDING	; Trigger ENV2 calculation
	bcf	STATUS,RP1	; Bank 0 Select
	return
	
	
; *******************************************************************
; *******************************************************************

	
; RAM Bank 0
CalculateEnv1
	bsf	STATUS,RP0	; RAM Bank 3
	bsf	STATUS,RP1
	
	btfsc	ENV_FLAGS,ENV1_PHASEA	; Decay/Sustain/Release phase?
	goto	CalculateEnv1_DS
	
	
	; Attack phase
	
	movlw	0x1F		; Set starting level to 0V (0x1FFE.00)
	movwf	ENV1_AMT_H
	movlw	0xFE
	movwf	ENV1_AMT_L
	clrf	ENV1_AMT_LL
	
	; ENV_PHASE_TIME * 13 => AARGB0,AARGB1,AARGB2
	bcf	STATUS,RP0	; RAM Bank 2
	movfw	vENV1_ATT_H
	bsf	STATUS,RP0	; RAM Bank 3
	movwf	AARGB0
	bcf	STATUS,RP0	; RAM Bank 2
	movfw	vENV1_ATT
	bsf	STATUS,RP0	; RAM Bank 3
	movwf	AARGB1
	movlw	D'13'
	movwf	BARGB0
	bcf	PCLATH,3	; Select program bank 2
	call	FXM1507U
	bsf	PCLATH,3	; Select program bank 3
	; AARGB1,AARGB2 => ENV1_STEPS_H,ENV1_STEPS_L
	movfw	AARGB1
	movwf	ENV1_STEPS_H
	movfw	AARGB2
	movwf	ENV1_STEPS_L
	
	; Work out DAC change ~= -(0x4185 * Level) >> 8
	; vENV1_ATL => BARGB0
	bcf	STATUS,RP0	; RAM Bank 2
	movfw	vENV1_ATL
	bsf	STATUS,RP0	; RAM Bank 3
	movwf	BARGB0
	; Set direction of steps
	bsf	ENV_FLAGS,ENV1_NEG	; If Level +ve, steps will be negative
	btfss	BARGB0,7
	goto	$+5
	bcf	ENV_FLAGS,ENV1_NEG	; Level -ve, so positive steps
	movfw	BARGB0
	sublw	0			; Flip sign of level
	movwf	BARGB0
	; (0x4185 * Level) => AARGB0,AARGB1,AARGB2
	movlw	0x41
	movwf	AARGB0
	movlw	0x85
	movwf	AARGB1
	bcf	PCLATH,3	; Select program bank 2
	call	FXM1507U
	bsf	PCLATH,3	; Select program bank 3
	
	; Special case that ENV_PHASE_TIME = 0?
	bcf	STATUS,RP0	; RAM Bank 2
	movfw	vENV1_ATT_H
	iorwf	vENV1_ATT,w
	bsf	STATUS,RP0	; RAM Bank 3
	jz	CalculateEnv1_T0	; Immediately add/subtract AARGB0,AARGB1,AARGB2 to/from envelope level
	
	; STEP_AMT = DAC_CHANGE * 0x100 / STEPS
	;            Remainder -> REM_AMT
	; AARGB0,AARGB1,AARGB2 / ENV1_STEPS_H,ENV1_STEPS_L
	call	CalculateEnv1_ChangeDivSteps
	
	goto	CalculateEnv1_End
	
	
CalculateEnv1_DS
	btfsc	ENV_FLAGS,ENV1_PHASEB	; Release phase?
	goto	CalculateEnv1_R
	
	
	; Decay/Sustain phase
	
	; ENV_PHASE_TIME * 13 => AARGB0,AARGB1,AARGB2
	bcf	STATUS,RP0	; RAM Bank 2
	movfw	vENV1_DEC_H
	bsf	STATUS,RP0	; RAM Bank 3
	movwf	AARGB0
	bcf	STATUS,RP0	; RAM Bank 2
	movfw	vENV1_DEC
	bsf	STATUS,RP0	; RAM Bank 3
	movwf	AARGB1
	movlw	D'13'
	movwf	BARGB0
	bcf	PCLATH,3	; Select program bank 2
	call	FXM1507U
	bsf	PCLATH,3	; Select program bank 3
	; AARGB1,AARGB2 => ENV1_STEPS_H,ENV1_STEPS_L
	movfw	AARGB1
	movwf	ENV1_STEPS_H
	movfw	AARGB2
	movwf	ENV1_STEPS_L
	
	; Work out new final DAC level ~= 0x1FFE - (0x4185 * Level) >> 8
	; (0x4185 * Level) => AARGB0,AARGB1,AARGB2
	bcf	STATUS,RP0	; RAM Bank 2
	movfw	vENV1_SUS
	bsf	STATUS,RP0	; RAM Bank 3
	movwf	BARGB0
	movlw	0x41
	movwf	AARGB0
	movlw	0x85
	movwf	AARGB1
	bcf	PCLATH,3	; Select program bank 2
	call	FXM1608S
	bsf	PCLATH,3	; Select program bank 3
	; 0x1FFE.00 - AARGB0,AARGB1.AARGB2
	movlw	0x1F
	movwf	REMB1
	movlw	0xFE
	movwf	BARGB0
	clrf	BARGB1
	SUB2424	BARGB1,AARGB2
	
	; Work out DAC change
	; REMB1,BARGB0.BARGB1 - ENV1_AMT_H,ENV1_AMT_L.ENV1_AMT_LL
	SUB2424 BARGB1,ENV1_AMT_LL
	
	; Set direction of steps, +ve DAC change => AARGB0,AARGB1.AARGB2
	call	CalculateEnv1_StepDir_FlipChange
	
	; Special case that ENV_PHASE_TIME = 0?
	bcf	STATUS,RP0	; RAM Bank 2
	movfw	vENV1_DEC_H
	iorwf	vENV1_DEC,w
	bsf	STATUS,RP0	; RAM Bank 3
	jz	CalculateEnv1_T0	; Immediately add/subtract AARGB0,AARGB1,AARGB2 to/from envelope level
	
	; STEP_AMT = DAC_CHANGE * 0x100 / STEPS
	;            Remainder -> REM_AMT
	; AARGB0,AARGB1,AARGB2 / ENV1_STEPS_H,ENV1_STEPS_L
	call	CalculateEnv1_ChangeDivSteps
	
	goto	CalculateEnv1_End
	
	
CalculateEnv1_R

	; Release phase
	
	; ENV_PHASE_TIME * 13 => AARGB0,AARGB1,AARGB2
	bcf	STATUS,RP0	; RAM Bank 2
	movfw	vENV1_REL_H
	bsf	STATUS,RP0	; RAM Bank 3
	movwf	AARGB0
	bcf	STATUS,RP0	; RAM Bank 2
	movfw	vENV1_REL
	bsf	STATUS,RP0	; RAM Bank 3
	movwf	AARGB1
	movlw	D'13'
	movwf	BARGB0
	bcf	PCLATH,3	; Select program bank 2
	call	FXM1507U
	bsf	PCLATH,3	; Select program bank 3
	; AARGB1,AARGB2 => ENV1_STEPS_H,ENV1_STEPS_L
	movfw	AARGB1
	movwf	ENV1_STEPS_H
	movfw	AARGB2
	movwf	ENV1_STEPS_L
	
	; Final DAC level needs to be 0x1FFE.00 => REMB1,BARGB0.BARGB1
	movlw	0x1F
	movwf	REMB1
	movlw	0xFE
	movwf	BARGB0
	clrf	BARGB1
	
	; Work out DAC change
	; REMB1,BARGB0.BARGB1 - ENV1_AMT_H,ENV1_AMT_L.ENV1_AMT_LL
	SUB2424 BARGB1,ENV1_AMT_LL
	
	; Set direction of steps, +ve DAC change => AARGB0,AARGB1.AARGB2
	call	CalculateEnv1_StepDir_FlipChange
	
	; Special case that ENV_PHASE_TIME = 0?
	bcf	STATUS,RP0	; RAM Bank 2
	movfw	vENV1_REL_H
	iorwf	vENV1_REL,w
	bsf	STATUS,RP0	; RAM Bank 3
	jz	CalculateEnv1_T0	; Immediately add/subtract AARGB0,AARGB1,AARGB2 to/from envelope level
	
	; STEP_AMT = DAC_CHANGE * 0x100 / STEPS
	;            Remainder -> REM_AMT
	; AARGB0,AARGB1,AARGB2 / ENV1_STEPS_H,ENV1_STEPS_L
	call	CalculateEnv1_ChangeDivSteps
	
	
CalculateEnv1_End
	bcf	STATUS,RP0	; RAM Bank 0
	bcf	STATUS,RP1
	bcf	MENU_FLAGS,ENV1_CALC_PENDING
	return
	
; *******************************************************************

CalculateEnv1_ChangeDivSteps
	; STEP_AMT = DAC_CHANGE * 0x100 / STEPS
	;            Remainder -> REM_AMT
	; AARGB0,AARGB1,AARGB2 / ENV1_STEPS_H,ENV1_STEPS_L
	movfw	ENV1_STEPS_H
	movwf	BARGB0
	movfw	ENV1_STEPS_L
	movwf	BARGB1
	bcf	PCLATH,3	; Select program bank 2
	call	FXD2416U
	bsf	PCLATH,3	; Select program bank 3
	; AARGB0,AARGB1,AARGB2 => ENV1_STEP_AMT_H,ENV1_STEP_AMT_L,ENV1_STEP_AMT_LL
	movfw	AARGB0
	movwf	ENV1_STEP_AMT_H
	movfw	AARGB1
	movwf	ENV1_STEP_AMT_L
	movfw	AARGB2
	movwf	ENV1_STEP_AMT_LL
	; REMB0,REMB1 => ENV1_REM_AMT_L,ENV1_REM_AMT_LL
	movfw	REMB0
	movwf	ENV1_REM_AMT_L
	movfw	REMB1
	movwf	ENV1_REM_AMT_LL
	
	return
	
; *******************************************************************

CalculateEnv1_StepDir_FlipChange
	; Set direction of steps, +ve DAC change => AARGB0,AARGB1.AARGB2
	btfsc	REMB1,7
	goto	$+9
	
	; +ve change, so REMB1,BARGB0.BARGB1 => AARGB0,AARGB1.AARGB2
	bcf	ENV_FLAGS,ENV1_NEG
	movfw	REMB1
	movwf	AARGB0
	movfw	BARGB0
	movwf	AARGB1
	movfw	BARGB1
	movwf	AARGB2
	return
	
	; -ve change, so 0 - REMB1,BARGB0.BARGB1 => AARGB0,AARGB1.AARGB2
	bsf	ENV_FLAGS,ENV1_NEG
	clrf AARGB0
	clrf AARGB1
	clrf AARGB2
	SUB2424	AARGB2,BARGB1
	return
	
; *******************************************************************
	
CalculateEnv1_T0
	; Immediately add/subtract AARGB0,AARGB1,AARGB2 to/from envelope level
	btfsc	ENV_FLAGS,ENV1_NEG	; Negative DAC change?
	goto	CalculateEnv1_T0_Neg
	ADD2424	ENV1_AMT_LL,AARGB2
	goto	CalculateEnv1_T0_1
CalculateEnv1_T0_Neg
	SUB2424	ENV1_AMT_LL,AARGB2
	
CalculateEnv1_T0_1
	; Create zero step so DAC gets updated during interrupt
	incf	ENV1_STEPS_L
	clrf	ENV1_STEP_AMT_H
	clrf	ENV1_STEP_AMT_L
	clrf	ENV1_STEP_AMT_LL
	clrf	ENV1_REM_AMT_L
	clrf	ENV1_REM_AMT_LL
	
	goto	CalculateEnv1_End
	
; *******************************************************************
; *******************************************************************
	
; RAM Bank 0
CalculateEnv2
	bsf	STATUS,RP0	; RAM Bank 3
	bsf	STATUS,RP1
	
	btfsc	ENV_FLAGS,ENV2_PHASEA	; Decay/Sustain/Release phase?
	goto	CalculateEnv2_DS
	
	
	; Attack phase
	
	movlw	0x1F		; Set starting level to 0V (0x1FFE.00)
	movwf	ENV2_AMT_H
	movlw	0xFE
	movwf	ENV2_AMT_L
	clrf	ENV2_AMT_LL
	
	; ENV_PHASE_TIME * 13 => AARGB0,AARGB1,AARGB2
	bcf	STATUS,RP0	; RAM Bank 2
	movfw	vENV2_ATT_H
	bsf	STATUS,RP0	; RAM Bank 3
	movwf	AARGB0
	bcf	STATUS,RP0	; RAM Bank 2
	movfw	vENV2_ATT
	bsf	STATUS,RP0	; RAM Bank 3
	movwf	AARGB1
	movlw	D'13'
	movwf	BARGB0
	bcf	PCLATH,3	; Select program bank 2
	call	FXM1507U
	bsf	PCLATH,3	; Select program bank 3
	; AARGB1,AARGB2 => ENV2_STEPS_H,ENV2_STEPS_L
	movfw	AARGB1
	movwf	ENV2_STEPS_H
	movfw	AARGB2
	movwf	ENV2_STEPS_L
	
	; Work out DAC change ~= -(0x4185 * Level) >> 8
	; vENV2_ATL => BARGB0
	bcf	STATUS,RP0	; RAM Bank 2
	movfw	vENV2_ATL
	bsf	STATUS,RP0	; RAM Bank 3
	movwf	BARGB0
	; Set direction of steps
	bsf	ENV_FLAGS,ENV2_NEG	; If Level +ve, steps will be negative
	btfss	BARGB0,7
	goto	$+5
	bcf	ENV_FLAGS,ENV2_NEG	; Level -ve, so positive steps
	movfw	BARGB0
	sublw	0			; Flip sign of level
	movwf	BARGB0
	; (0x4185 * Level) => AARGB0,AARGB1,AARGB2
	movlw	0x41
	movwf	AARGB0
	movlw	0x85
	movwf	AARGB1
	bcf	PCLATH,3	; Select program bank 2
	call	FXM1507U
	bsf	PCLATH,3	; Select program bank 3
	
	; Special case that ENV_PHASE_TIME = 0?
	bcf	STATUS,RP0	; RAM Bank 2
	movfw	vENV2_ATT_H
	iorwf	vENV2_ATT,w
	bsf	STATUS,RP0	; RAM Bank 3
	jz	CalculateEnv2_T0	; Immediately add/subtract AARGB0,AARGB1,AARGB2 to/from envelope level
	
	; STEP_AMT = DAC_CHANGE * 0x100 / STEPS
	;            Remainder -> REM_AMT
	; AARGB0,AARGB1,AARGB2 / ENV2_STEPS_H,ENV2_STEPS_L
	call	CalculateEnv2_ChangeDivSteps
	
	goto	CalculateEnv2_End
	
	
CalculateEnv2_DS
	btfsc	ENV_FLAGS,ENV2_PHASEB	; Release phase?
	goto	CalculateEnv2_R
	
	
	; Decay/Sustain phase
	
	; ENV_PHASE_TIME * 13 => AARGB0,AARGB1,AARGB2
	bcf	STATUS,RP0	; RAM Bank 2
	movfw	vENV2_DEC_H
	bsf	STATUS,RP0	; RAM Bank 3
	movwf	AARGB0
	bcf	STATUS,RP0	; RAM Bank 2
	movfw	vENV2_DEC
	bsf	STATUS,RP0	; RAM Bank 3
	movwf	AARGB1
	movlw	D'13'
	movwf	BARGB0
	bcf	PCLATH,3	; Select program bank 2
	call	FXM1507U
	bsf	PCLATH,3	; Select program bank 3
	; AARGB1,AARGB2 => ENV2_STEPS_H,ENV2_STEPS_L
	movfw	AARGB1
	movwf	ENV2_STEPS_H
	movfw	AARGB2
	movwf	ENV2_STEPS_L
	
	; Work out new final DAC level ~= 0x1FFE - (0x4185 * Level) >> 8
	; (0x4185 * Level) => AARGB0,AARGB1,AARGB2
	bcf	STATUS,RP0	; RAM Bank 2
	movfw	vENV2_SUS
	bsf	STATUS,RP0	; RAM Bank 3
	movwf	BARGB0
	movlw	0x41
	movwf	AARGB0
	movlw	0x85
	movwf	AARGB1
	bcf	PCLATH,3	; Select program bank 2
	call	FXM1608S
	bsf	PCLATH,3	; Select program bank 3
	; 0x1FFE.00 - AARGB0,AARGB1.AARGB2
	movlw	0x1F
	movwf	REMB1
	movlw	0xFE
	movwf	BARGB0
	clrf	BARGB1
	SUB2424	BARGB1,AARGB2
	
	; Work out DAC change
	; REMB1,BARGB0.BARGB1 - ENV2_AMT_H,ENV2_AMT_L.ENV2_AMT_LL
	SUB2424 BARGB1,ENV2_AMT_LL
	
	; Set direction of steps, +ve DAC change => AARGB0,AARGB1.AARGB2
	call	CalculateEnv2_StepDir_FlipChange
	
	; Special case that ENV_PHASE_TIME = 0?
	bcf	STATUS,RP0	; RAM Bank 2
	movfw	vENV2_DEC_H
	iorwf	vENV2_DEC,w
	bsf	STATUS,RP0	; RAM Bank 3
	jz	CalculateEnv2_T0	; Immediately add/subtract AARGB0,AARGB1,AARGB2 to/from envelope level
	
	; STEP_AMT = DAC_CHANGE * 0x100 / STEPS
	;            Remainder -> REM_AMT
	; AARGB0,AARGB1,AARGB2 / ENV2_STEPS_H,ENV2_STEPS_L
	call	CalculateEnv2_ChangeDivSteps
	
	goto	CalculateEnv2_End
	
	
CalculateEnv2_R

	; Release phase
	
	; ENV_PHASE_TIME * 13 => AARGB0,AARGB1,AARGB2
	bcf	STATUS,RP0	; RAM Bank 2
	movfw	vENV2_REL_H
	bsf	STATUS,RP0	; RAM Bank 3
	movwf	AARGB0
	bcf	STATUS,RP0	; RAM Bank 2
	movfw	vENV2_REL
	bsf	STATUS,RP0	; RAM Bank 3
	movwf	AARGB1
	movlw	D'13'
	movwf	BARGB0
	bcf	PCLATH,3	; Select program bank 2
	call	FXM1507U
	bsf	PCLATH,3	; Select program bank 3
	; AARGB1,AARGB2 => ENV2_STEPS_H,ENV2_STEPS_L
	movfw	AARGB1
	movwf	ENV2_STEPS_H
	movfw	AARGB2
	movwf	ENV2_STEPS_L
	
	; Final DAC level needs to be 0x1FFE.00 => REMB1,BARGB0.BARGB1
	movlw	0x1F
	movwf	REMB1
	movlw	0xFE
	movwf	BARGB0
	clrf	BARGB1
	
	; Work out DAC change
	; REMB1,BARGB0.BARGB1 - ENV2_AMT_H,ENV2_AMT_L.ENV2_AMT_LL
	SUB2424 BARGB1,ENV2_AMT_LL
	
	; Set direction of steps, +ve DAC change => AARGB0,AARGB1.AARGB2
	call	CalculateEnv2_StepDir_FlipChange
	
	; Special case that ENV_PHASE_TIME = 0?
	bcf	STATUS,RP0	; RAM Bank 2
	movfw	vENV2_REL_H
	iorwf	vENV2_REL,w
	bsf	STATUS,RP0	; RAM Bank 3
	jz	CalculateEnv2_T0	; Immediately add/subtract AARGB0,AARGB1,AARGB2 to/from envelope level
	
	; STEP_AMT = DAC_CHANGE * 0x100 / STEPS
	;            Remainder -> REM_AMT
	; AARGB0,AARGB1,AARGB2 / ENV2_STEPS_H,ENV2_STEPS_L
	call	CalculateEnv2_ChangeDivSteps
	
	
CalculateEnv2_End
	bcf	STATUS,RP0	; RAM Bank 0
	bcf	STATUS,RP1
	bcf	MENU_FLAGS,ENV2_CALC_PENDING
	return
	
; *******************************************************************

CalculateEnv2_ChangeDivSteps
	; STEP_AMT = DAC_CHANGE * 0x100 / STEPS
	;            Remainder -> REM_AMT
	; AARGB0,AARGB1,AARGB2 / ENV2_STEPS_H,ENV2_STEPS_L
	movfw	ENV2_STEPS_H
	movwf	BARGB0
	movfw	ENV2_STEPS_L
	movwf	BARGB1
	bcf	PCLATH,3	; Select program bank 2
	call	FXD2416U
	bsf	PCLATH,3	; Select program bank 3
	; AARGB0,AARGB1,AARGB2 => ENV2_STEP_AMT_H,ENV2_STEP_AMT_L,ENV2_STEP_AMT_LL
	movfw	AARGB0
	movwf	ENV2_STEP_AMT_H
	movfw	AARGB1
	movwf	ENV2_STEP_AMT_L
	movfw	AARGB2
	movwf	ENV2_STEP_AMT_LL
	; REMB0,REMB1 => ENV2_REM_AMT_L,ENV2_REM_AMT_LL
	movfw	REMB0
	movwf	ENV2_REM_AMT_L
	movfw	REMB1
	movwf	ENV2_REM_AMT_LL
	
	return
	
; *******************************************************************

CalculateEnv2_StepDir_FlipChange
	; Set direction of steps, +ve DAC change => AARGB0,AARGB1.AARGB2
	btfsc	REMB1,7
	goto	$+9
	
	; +ve change, so REMB1,BARGB0.BARGB1 => AARGB0,AARGB1.AARGB2
	bcf	ENV_FLAGS,ENV2_NEG
	movfw	REMB1
	movwf	AARGB0
	movfw	BARGB0
	movwf	AARGB1
	movfw	BARGB1
	movwf	AARGB2
	return
	
	; -ve change, so 0 - REMB1,BARGB0.BARGB1 => AARGB0,AARGB1.AARGB2
	bsf	ENV_FLAGS,ENV2_NEG
	clrf AARGB0
	clrf AARGB1
	clrf AARGB2
	SUB2424	AARGB2,BARGB1
	return
	
; *******************************************************************
	
CalculateEnv2_T0
	; Immediately add/subtract AARGB0,AARGB1,AARGB2 to/from envelope level
	btfsc	ENV_FLAGS,ENV2_NEG	; Negative DAC change?
	goto	CalculateEnv2_T0_Neg
	ADD2424	ENV2_AMT_LL,AARGB2
	goto	CalculateEnv2_T0_1
CalculateEnv2_T0_Neg
	SUB2424	ENV2_AMT_LL,AARGB2
	
CalculateEnv2_T0_1
	; Create zero step so DAC gets updated during interrupt
	incf	ENV2_STEPS_L
	clrf	ENV2_STEP_AMT_H
	clrf	ENV2_STEP_AMT_L
	clrf	ENV2_STEP_AMT_LL
	clrf	ENV2_REM_AMT_L
	clrf	ENV2_REM_AMT_LL
	
	goto	CalculateEnv2_End
	
	
; *******************************************************************
; *******************************************************************
; *******************************************************************


; RAM Bank 0
; MTEMP2=Controller Number
; MTEMP3=Controller Value
ChangeParamViaMIDI
	movlw	high ChangeParamViaMIDI_JumpTable
	movwf	PCLATH
	movlw	low ChangeParamViaMIDI_JumpTable
	addwf	MTEMP2,w
	snc
	incf	PCLATH
	movwf	PCL
	
ChangeParamViaMIDI_JumpTable
	return			; CC0
	return			; CC1
	return			; CC2
	return			; CC3
	return			; CC4
	goto Change_vPORTA_T1	; CC5
	return			; CC6
	goto Change_vVELOCITY	; CC7
	goto Change_vOCTAVE_2	; CC8
	return			; CC9
	return			; CC10
	return			; CC11
	return			; CC12
	return			; CC13
	return			; CC14
	return			; CC15
	return			; CC16
	return			; CC17
	return			; CC18
	goto Change_vLFO1_TIM	; CC19
	goto Change_vLFO1_WAV	; CC20
	return			; CC21
	goto Change_vLFO1_AMT	; CC22
	goto Change_vLFO2_TIM	; CC23
	goto Change_vLFO2_WAV	; CC24
	goto Change_vLFO2_AMT	; CC25
	goto Change_vPORTA_T2	; CC26
	goto Change_vPORTA_T3	; CC27
	return			; CC28
	goto Change_vP_BEND_1	; CC29
	return			; CC30
	return			; CC31
	return			; CC32
	goto Change_vP_BEND_3	; CC33
	return			; CC34
	return			; CC35
	goto Change_vENV1_DEC	; CC36
	goto Change_vENV1_SUS	; CC37
	goto Change_vENV2_ATT	; CC38
	goto Change_vENV2_DEC	; CC39
	goto Change_vENV2_SUS	; CC40
	goto Change_vENV2_REL	; CC41
	goto Change_vENV2_ATL	; CC42
	goto Change_vOCTAVE_3	; CC43
	goto Change_vPORTA	; CC44
	return			; CC45
	goto Change_vL_RETRIG	; CC46
	return			; CC47
	return			; CC48
	return			; CC49
	return			; CC50
	return			; CC51
	return			; CC52
	return			; CC53
	return			; CC54
	return			; CC55
	return			; CC56
	return			; CC57
	return			; CC58
	return			; CC59
	return			; CC60
	return			; CC61
	return			; CC62
	return			; CC63
	return			; CC64
	return			; CC65
	return			; CC66
	return			; CC67
	return			; CC68
	return			; CC69
	goto Change_vP_BEND_2	; CC70
	return			; CC71
	goto Change_vENV1_REL	; CC72
	goto Change_vENV1_ATT	; CC73
	goto Change_vENV1_ATL	; CC74
	return			; CC75
	return			; CC76
	return			; CC77
	goto Change_vOCTAVE_1	; CC78
	goto Change_vN_O_GATE	; CC79
	return			; CC80
	return			; CC81
	return			; CC82
	return			; CC83
	return			; CC84
	return			; CC85
	return			; CC86
	return			; CC87
	return			; CC88
	return			; CC89
	return			; CC90
	return			; CC91
	return			; CC92
	return			; CC93
	return			; CC94
	return			; CC95
	return			; CC96
	return			; CC97
	return			; CC98
	return			; CC99
	return			; CC100
	goto Change_vAMOUNT_CV1	; CC101
	goto Change_vAMOUNT_CV2	; CC102
	goto Change_vAMOUNT_CV3	; CC103
	goto Change_vAMOUNT_CV4	; CC104
	goto Change_vAMOUNT_CV5	; CC105
	goto Change_vAMOUNT_CV6	; CC106
	goto Change_vAMOUNT_CV7	; CC107
	goto Change_vAMOUNT_CV8	; CC108
	goto Change_vAMOUNT_CV9	; CC109
	goto Change_vAMOUNT_CV10 ; CC110
	goto Change_vAMOUNT_CV11 ; CC111
	goto Change_vAMOUNT_CV12 ; CC112
	goto Change_vAMOUNT_CV13 ; CC113
	goto Change_vAMOUNT_CV14 ; CC114
	goto Change_vAMOUNT_CV15 ; CC115
	return			; CC116
	return			; CC117
	return			; CC118
	return			; CC119
	return			; CC120
	return			; CC121
	return			; CC122
	return			; CC123
	return			; CC124
	return			; CC125
	return			; CC126
	return			; CC127
	
; *******************************************************************
	
; RAM Bank 0
; MTEMP2=Controller Number
; MTEMP3=Controller Value
Change_vL_RETRIG
	; 0~Off, >0~On
	movfw	MTEMP3
	movlw	0
	sz
	movlw	1
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vL_RETRIG
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay

Change_vN_O_GATE
	; 0~Off, 1~0.01s, 127~1.27s
	; MSB unaffected
	movfw	MTEMP3
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vN_O_GATE
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vVELOCITY
	; 0~Off, 1~0.06V, >=125~7.50V
	; Double CC value
	movfw	MTEMP3
	sublw	D'125'
	jc	$+3
	movlw	D'125'
	goto	$+2
	movfw	MTEMP3
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vVELOCITY
	bcf	STATUS,C
	rlf	vVELOCITY
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vPORTA
	; 0~0, 4~1, 1~2, 3~3, 2~4
	; Map values
	movfw	MTEMP3
	jz	Change_vPORTA_0
	decf	MTEMP3,w
	jz	Change_vPORTA_2
	movwf	MTEMP1
	decf	MTEMP1
	jz	Change_vPORTA_4
	decf	MTEMP1
	jz	Change_vPORTA_3
	decf	MTEMP1
	jz	Change_vPORTA_1
	return
Change_vPORTA_0
	movlw	0
	goto	$+8
Change_vPORTA_1
	movlw	1
	goto	$+6
Change_vPORTA_2
	movlw	2
	goto	$+4
Change_vPORTA_3
	movlw	3
	goto	$+2
Change_vPORTA_4
	movlw	4
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vPORTA
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vPORTA_T1
	; 0~0.04s, 127~1.31s
	; CC value+4, MSB unaffected
	movfw	MTEMP3
	addlw	4
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vPORTA_T1
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vPORTA_T2
	; 0~0.04s, 127~1.31s
	; CC value+4, MSB unaffected
	movfw	MTEMP3
	addlw	4
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vPORTA_T2
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vPORTA_T3
	; 0~0.04s, 127~1.31s
	; CC value+4, MSB unaffected
	movfw	MTEMP3
	addlw	4
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vPORTA_T3
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vP_BEND_1
	; 0~1, 127~128
	; CC value+1
	movfw	MTEMP3
	addlw	1
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vP_BEND_1
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vP_BEND_2
	; 0~1, 127~128
	; CC value+1
	movfw	MTEMP3
	addlw	1
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vP_BEND_2
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vP_BEND_3
	; 0~1, 127~128
	; CC value+1
	movfw	MTEMP3
	addlw	1
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vP_BEND_3
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vOCTAVE_1
	; 0~-5, 12~-4, 24~-3, 36~-2, 48~-1,
	; 60~0, 72~1, 84~2, 96~3, 108~4, 120~5
	; CC value/12 - 5
	; Only respond to exact values
	bsf	STATUS,RP0	; RAM Bank 1
	movfw	MTEMP3
	movwf	AARGB0
	movlw	D'12'
	movwf	BARGB0
	call	FXD0707U
	movfw	REMB0
	bcf	STATUS,RP0	; RAM Bank 0
	rnz
	bsf	STATUS,RP0	; RAM Bank 1
	movfw	AARGB0
	bcf	STATUS,RP0	; RAM Bank 2
	bsf	STATUS,RP1
	addlw	0xFB		; -5
	movwf	vOCTAVE_1
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vOCTAVE_2
	; 0~-5, 12~-4, 24~-3, 36~-2, 48~-1,
	; 60~0, 72~1, 84~2, 96~3, 108~4, 120~5
	; CC value/12 - 5
	; Only respond to exact values
	bsf	STATUS,RP0	; RAM Bank 1
	movfw	MTEMP3
	movwf	AARGB0
	movlw	D'12'
	movwf	BARGB0
	call	FXD0707U
	movfw	REMB0
	bcf	STATUS,RP0	; RAM Bank 0
	rnz
	bsf	STATUS,RP0	; RAM Bank 1
	movfw	AARGB0
	bcf	STATUS,RP0	; RAM Bank 2
	bsf	STATUS,RP1
	addlw	0xFB		; -5
	movwf	vOCTAVE_2
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vOCTAVE_3
	; 0~-5, 12~-4, 24~-3, 36~-2, 48~-1,
	; 60~0, 72~1, 84~2, 96~3, 108~4, 120~5
	; CC value/12 - 5
	; Only respond to exact values
	bsf	STATUS,RP0	; RAM Bank 1
	movfw	MTEMP3
	movwf	AARGB0
	movlw	D'12'
	movwf	BARGB0
	call	FXD0707U
	movfw	REMB0
	bcf	STATUS,RP0	; RAM Bank 0
	rnz
	bsf	STATUS,RP0	; RAM Bank 1
	movfw	AARGB0
	bcf	STATUS,RP0	; RAM Bank 2
	bsf	STATUS,RP1
	addlw	0xFB		; -5
	movwf	vOCTAVE_3
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vLFO1_TIM
	; 0~0.01s, 127~1.28s
	; CC value+1, MSB unaffected
	movfw	MTEMP3
	addlw	1
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vLFO1_TIM
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vLFO1_WAV
	; 0~5, 1~2, 2~1, 3~3, 4~4, 5~0
	; Map values
	movfw	MTEMP3
	jz	Change_vLFO1_WAV_5
	decf	MTEMP3,w
	jz	Change_vLFO1_WAV_2
	movwf	MTEMP1
	decf	MTEMP1
	jz	Change_vLFO1_WAV_1
	decf	MTEMP1
	jz	Change_vLFO1_WAV_3
	decf	MTEMP1
	jz	Change_vLFO1_WAV_4
	decf	MTEMP1
	jz	Change_vLFO1_WAV_0
	return
Change_vLFO1_WAV_0
	movlw	0
	goto	Change_vLFO1_WAV_X
Change_vLFO1_WAV_1
	movlw	1
	goto	Change_vLFO1_WAV_X
Change_vLFO1_WAV_2
	movlw	2
	goto	Change_vLFO1_WAV_X
Change_vLFO1_WAV_3
	movlw	3
	goto	Change_vLFO1_WAV_X
Change_vLFO1_WAV_4
	movlw	4
	goto	Change_vLFO1_WAV_X
Change_vLFO1_WAV_5
	movlw	5
Change_vLFO1_WAV_X
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vLFO1_WAV
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vLFO1_AMT
	; 0~Off, 1~0.06V, >=125~7.50V
	; Double CC value
	movfw	MTEMP3
	sublw	D'125'
	jc	$+3
	movlw	D'125'
	goto	$+2
	movfw	MTEMP3
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vLFO1_AMT
	bcf	STATUS,C
	rlf	vLFO1_AMT
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vLFO2_TIM
	; 0~0.01s, 127~1.28s
	; CC value+1, MSB unaffected
	movfw	MTEMP3
	addlw	1
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vLFO2_TIM
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vLFO2_WAV
	; 0~5, 1~2, 2~1, 3~3, 4~4, 5~0
	; Map values
	movfw	MTEMP3
	jz	Change_vLFO2_WAV_5
	decf	MTEMP3,w
	jz	Change_vLFO2_WAV_2
	movwf	MTEMP1
	decf	MTEMP1
	jz	Change_vLFO2_WAV_1
	decf	MTEMP1
	jz	Change_vLFO2_WAV_3
	decf	MTEMP1
	jz	Change_vLFO2_WAV_4
	decf	MTEMP1
	jz	Change_vLFO2_WAV_0
	return
Change_vLFO2_WAV_0
	movlw	0
	goto	Change_vLFO2_WAV_X
Change_vLFO2_WAV_1
	movlw	1
	goto	Change_vLFO2_WAV_X
Change_vLFO2_WAV_2
	movlw	2
	goto	Change_vLFO2_WAV_X
Change_vLFO2_WAV_3
	movlw	3
	goto	Change_vLFO2_WAV_X
Change_vLFO2_WAV_4
	movlw	4
	goto	Change_vLFO2_WAV_X
Change_vLFO2_WAV_5
	movlw	5
Change_vLFO2_WAV_X
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vLFO2_WAV
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vLFO2_AMT
	; 0~Off, 1~0.06V, >=125~7.50V
	; Double CC value
	movfw	MTEMP3
	sublw	D'125'
	jc	$+3
	movlw	D'125'
	goto	$+2
	movfw	MTEMP3
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vLFO2_AMT
	bcf	STATUS,C
	rlf	vLFO2_AMT
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vENV1_ATT
	; 0~0.00s, 127~1.27s
	; MSB unaffected
	movfw	MTEMP3
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vENV1_ATT
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vENV1_ATL
	; 0~0.00V, >=125~7.50V
	; Positive range only
	movfw	MTEMP3
	sublw	D'125'
	jc	$+3
	movlw	D'125'
	goto	$+2
	movfw	MTEMP3
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vENV1_ATL
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vENV1_DEC
	; 0~0.00s, 127~1.27s
	; MSB unaffected
	movfw	MTEMP3
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vENV1_DEC
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vENV1_SUS
	; 0~0.00V, >=125~7.50V
	; Positive range only
	movfw	MTEMP3
	sublw	D'125'
	jc	$+3
	movlw	D'125'
	goto	$+2
	movfw	MTEMP3
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vENV1_SUS
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vENV1_REL
	; 0~0.00s, 127~1.27s
	; MSB unaffected
	movfw	MTEMP3
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vENV1_REL
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vENV2_ATT
	; 0~0.00s, 127~1.27s
	; MSB unaffected
	movfw	MTEMP3
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vENV2_ATT
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vENV2_ATL
	; 0~0.00V, >=125~7.50V
	; Positive range only
	movfw	MTEMP3
	sublw	D'125'
	jc	$+3
	movlw	D'125'
	goto	$+2
	movfw	MTEMP3
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vENV2_ATL
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vENV2_DEC
	; 0~0.00s, 127~1.27s
	; MSB unaffected
	movfw	MTEMP3
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vENV2_DEC
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vENV2_SUS
	; 0~0.00V, >=125~7.50V
	; Positive range only
	movfw	MTEMP3
	sublw	D'125'
	jc	$+3
	movlw	D'125'
	goto	$+2
	movfw	MTEMP3
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vENV2_SUS
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vENV2_REL
	; 0~0.00s, 127~1.27s
	; MSB unaffected
	movfw	MTEMP3
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vENV2_REL
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vAMOUNT_CV1
	; 0~0.00V, >=125~7.50V
	; Positive range only
	movfw	MTEMP3
	sublw	D'125'
	jc	$+3
	movlw	D'125'
	goto	$+2
	movfw	MTEMP3
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vAMOUNT_CV1
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vAMOUNT_CV2
	; 0~0.00V, >=125~7.50V
	; Positive range only
	movfw	MTEMP3
	sublw	D'125'
	jc	$+3
	movlw	D'125'
	goto	$+2
	movfw	MTEMP3
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vAMOUNT_CV2
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vAMOUNT_CV3
	; 0~0.00V, >=125~7.50V
	; Positive range only
	movfw	MTEMP3
	sublw	D'125'
	jc	$+3
	movlw	D'125'
	goto	$+2
	movfw	MTEMP3
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vAMOUNT_CV3
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vAMOUNT_CV4
	; 0~0.00V, >=125~7.50V
	; Positive range only
	movfw	MTEMP3
	sublw	D'125'
	jc	$+3
	movlw	D'125'
	goto	$+2
	movfw	MTEMP3
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vAMOUNT_CV4
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vAMOUNT_CV5
	; 0~0.00V, >=125~7.50V
	; Positive range only
	movfw	MTEMP3
	sublw	D'125'
	jc	$+3
	movlw	D'125'
	goto	$+2
	movfw	MTEMP3
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vAMOUNT_CV5
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vAMOUNT_CV6
	; 0~0.00V, >=125~7.50V
	; Positive range only
	movfw	MTEMP3
	sublw	D'125'
	jc	$+3
	movlw	D'125'
	goto	$+2
	movfw	MTEMP3
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vAMOUNT_CV6
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vAMOUNT_CV7
	; 0~0.00V, >=125~7.50V
	; Positive range only
	movfw	MTEMP3
	sublw	D'125'
	jc	$+3
	movlw	D'125'
	goto	$+2
	movfw	MTEMP3
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vAMOUNT_CV7
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vAMOUNT_CV8
	; 0~0.00V, >=125~7.50V
	; Positive range only
	movfw	MTEMP3
	sublw	D'125'
	jc	$+3
	movlw	D'125'
	goto	$+2
	movfw	MTEMP3
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vAMOUNT_CV8
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vAMOUNT_CV9
	; 0~0.00V, >=125~7.50V
	; Positive range only
	movfw	MTEMP3
	sublw	D'125'
	jc	$+3
	movlw	D'125'
	goto	$+2
	movfw	MTEMP3
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vAMOUNT_CV9
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vAMOUNT_CV10
	; 0~0.00V, >=125~7.50V
	; Positive range only
	movfw	MTEMP3
	sublw	D'125'
	jc	$+3
	movlw	D'125'
	goto	$+2
	movfw	MTEMP3
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vAMOUNT_CV10
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vAMOUNT_CV11
	; 0~0.00V, >=125~7.50V
	; Positive range only
	movfw	MTEMP3
	sublw	D'125'
	jc	$+3
	movlw	D'125'
	goto	$+2
	movfw	MTEMP3
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vAMOUNT_CV11
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vAMOUNT_CV12
	; 0~0.00V, >=125~7.50V
	; Positive range only
	movfw	MTEMP3
	sublw	D'125'
	jc	$+3
	movlw	D'125'
	goto	$+2
	movfw	MTEMP3
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vAMOUNT_CV12
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vAMOUNT_CV13
	; 0~0.00V, >=125~7.50V
	; Positive range only
	movfw	MTEMP3
	sublw	D'125'
	jc	$+3
	movlw	D'125'
	goto	$+2
	movfw	MTEMP3
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vAMOUNT_CV13
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vAMOUNT_CV14
	; 0~0.00V, >=125~7.50V
	; Positive range only
	movfw	MTEMP3
	sublw	D'125'
	jc	$+3
	movlw	D'125'
	goto	$+2
	movfw	MTEMP3
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vAMOUNT_CV14
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
Change_vAMOUNT_CV15
	; 0~0.00V, >=125~7.50V
	; Positive range only
	movfw	MTEMP3
	sublw	D'125'
	jc	$+3
	movlw	D'125'
	goto	$+2
	movfw	MTEMP3
	bsf	STATUS,RP1	; RAM Bank 2
	movwf	vAMOUNT_CV15
	bcf	STATUS,RP1	; RAM Bank 0
	goto	Change_UpdateDisplay
	
; *******************************************************************

Change_UpdateDisplay
	bcf	PCLATH,3	; Select program bank 0
	bcf	PCLATH,4
	bcf	INTCON,GIE	; Disable interrupts while updating display
	call	UpdateDisplay_ValueOnly
	bsf	INTCON,GIE	; Enable interrupts again
	bsf	PCLATH,3	; Select program bank 3
	bsf	PCLATH,4
	return


; *******************************************************************
; *******************************************************************
; *******************************************************************

; More maths Functions
#include <maths2.inc>

;#######################################################################


	END                       ; directive 'end of program'


